/*
Analyze coverage file
*/

#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <map>
#include <string>
#include <cmath>
#include "mergejunc.h"
#include "rangeset.h"
#include "commontype.h"

using namespace std;

/* Min number of junctions supporting this position */
int CMIN_JC=1;

/* Min number of samples supporting this position */
int CMIN_JSAMPLE=2;


/* Map variable to mark whether we observe a zero coverage region around a junction point */
AllChrJdir JZeroMark;

/* Map variable to store candidate coverage cuts */ 
AllChrJdir JCandidateCut;

typedef map<long,int> ChrCvg;
typedef map<string, ChrCvg> AllChrCvg;

void printcvgmsg(){
  cout<<"Usage: manalyzecvg {Options} <WIG file 1> <WIG file 2> ...\n";
}

/* Calculate the minimum distance from pos to any key value in the cvg */
long minDistToMap(long pos, map<long,int>& cvg, long& closestpoint){
  long x=pos;
  map<long,int>::iterator ihow=cvg.upper_bound(x);
  if (ihow==cvg.end()){
    closestpoint=cvg.rbegin()->first;
    return abs(double(x+1-(cvg.rbegin()->first)));
  }
  closestpoint=ihow->first;
  long r=abs(double(x-ihow->first));
  if(ihow!=cvg.begin()){
    ihow--;//should point to 0
    long r2=abs(double(x+1-ihow->first));
    if(r2<r){
      r=r2;
      closestpoint=ihow->first;
    }
  }   
  return r;
}

/* Calculate the minimum distance from pos to the closest point in cvg, with search direction:
-1, search left; 1, search right. 

*/
long minJuncInMap(long pos, map<long,int>& cvg, long& closestpoint,int dir){
  long x=pos;
  map<long,int>::iterator ihow=cvg.upper_bound(x);
  if(dir>0){
    if (ihow==cvg.end()){
      closestpoint=-1;
      return -1;
    }
    closestpoint=ihow->first;
    return abs(double(x+1-(ihow->first)));
  }

  while(ihow!=cvg.begin() && ihow->first>pos){
    ihow--;
  }
  if(ihow->first>pos){
    closestpoint=-1;
    return -1;
  }
  closestpoint=ihow->first;
  long r=abs(double(x-ihow->first));
  return r;
}


/* Process zero region, replaced by processzeropoint */
int processZeroRegion(long zerostart, long zeroend, string chr){
  long MIN_DIST_2EXISTING_JUNC=100;
  long MIN_ZREGION_LENGTH=100;
  if(zeroend-zerostart<MIN_ZREGION_LENGTH)return -1;
  //output zero zones
  //check if the zero boundaries are close to junctions?
  long p_j_start,p_j_end; //records the closest junction corresponding to the zero boundary
  long t1=minDistToMap(zerostart,JDIR[chr],p_j_start);
  long t2=minDistToMap(zeroend,JDIR[chr],p_j_end);

  if(t1>MIN_DIST_2EXISTING_JUNC){
    cout<<chr<<"\t"<<zerostart<<"\t"<<zerostart<<"\t+"<<endl;
  }else{
    JZeroMark[chr][p_j_start]++;
  }
  if(t1>MIN_DIST_2EXISTING_JUNC){
    cout<<chr<<"\t"<<zeroend<<"\t"<<zeroend<<"\t-"<<endl;
  }else{
    JZeroMark[chr][p_j_end]++;
  }

  return 0;
}

/* processzeropoint: for each candidate point with 0 coverage, check if they are close to existing junction annotations. 
dir: the direction of the zero point, -1 if it's start of the region, 1 if the end of the region*/
int processZeroPoint(long zeropt, string chr,int dir){
  long MIN_DIST_2EXISTING_JUNC=900;
  long MIN_DIST_2CLOSESTJUNC=5;
  //check if the zero boundaries are close to junctions?
  long p_j_start; //records the closest junction corresponding to the zero boundary
  long t0=minDistToMap(zeropt,JDIR[chr],p_j_start); //the distance to its nearest junction point without direction
  long t1=minJuncInMap(zeropt,JDIR[chr],p_j_start,dir); //the distance to its nearest junction point along direction

  int objjdir=0;
  if(p_j_start!=-1) objjdir=JDIR[chr][p_j_start];
  //cout<<"Pos:"<<zeropt<<",dir:"<<dir<<",objdir:"<<objjdir<<",t1="<<t1<<endl;
  if((t1>MIN_DIST_2EXISTING_JUNC || objjdir*dir>=0) && t0>MIN_DIST_2CLOSESTJUNC){
    //cout<<chr<<"\t"<<zeropt<<"\t"<<zeropt<<"\t+"<<endl;
    JCandidateCut[chr][zeropt]=0;
  }else{
    if(p_j_start!=-1)
      JZeroMark[chr][p_j_start]++;
  }


  return 0;
}


/* Scan the coverage file */
int scanCvg(string filename ){
  int MAX_JUMP_CVG=5;
  ifstream ifs(filename.c_str());
  if(!ifs.is_open()){
    cerr<<"Error opening file "<<filename<<endl;
    return -1;
  }
  string prevchr="";
 
  long prevendpos=-1;
  long zrstart=0,zrend=-1;
  while(true){
    string oneline;
    getline(ifs,oneline);
    if(ifs.eof())break;
    stringstream ss(oneline);
    string chrname;
    long startpos,endpos;
    int count;
    ss>>chrname;
    if(chrname=="track")continue;
    ss>>startpos; ss>>endpos;ss>>count;

    if(startpos!=prevendpos && prevchr==chrname){
      //encounters one zero region
      //processZeroRegion(prevendpos,startpos,chrname);
      if(zrstart==-1) {
        zrstart=prevendpos;
        processZeroPoint(zrstart,chrname, -1);
      }
      zrend=startpos;
    }
    if(count==0){
      if(zrstart==-1) {
          zrstart=startpos;
          processZeroPoint(zrstart,chrname,-1);
      }
      zrend=endpos;
    }else{
      if(count>MAX_JUMP_CVG){
        if(zrstart!=-1){
          //processZeroRegion(zrstart,zrend,chrname);
          //processZeroPoint(zrstart,chrname);
          zrstart=-1;
        }
        if(zrend!=-1) {
          processZeroPoint(zrend,chrname,1);
          zrend=-1;
        }
      }
    }
    
    //process right now, don't save anymore (it takes too much memory)
    //all[chrname][startpos]=count;
    //all[chrname][endpos]=0;
    prevendpos=endpos;
    prevchr=chrname;
  }
  
  ifs.close();
  return 0;
}

/* Scan the coverage file, calculate the Haar score */
int scanCvgandHaarScore(string filename, AllChrCvg& all, int windowsize){
  ifstream ifs(filename.c_str());
  if(!ifs.is_open()){
    cerr<<"Error opening file "<<filename<<endl;
    return -1;
  }
  string prevchr="";
 
  long prevendpos=-1;
  long zrstart=-1,zrend=-1;
  ChrCvg* currentChr;
  ChrCvg::iterator kleft,kright,kcr;
  while(true){
    string oneline;
    getline(ifs,oneline);
    if(ifs.eof())break;
    stringstream ss(oneline);
    string chrname;
    long startpos,endpos;
    int count;
    ss>>chrname;
    if(chrname=="track")continue;
    ss>>startpos; ss>>endpos;ss>>count;
    
    if(prevchr!=chrname){
      currentChr=&(all[chrname]);
      //point kleft and kright
      kleft=currentChr->upper_bound(startpos-windowsize);
      kright=currentChr->upper_bound(startpos+windowsize);
    }
    while(kleft!=currentChr->end() && kleft->first<startpos-windowsize)kleft++;
    while(kright!=currentChr->end() && kright->first<startpos+windowsize)kright++;
    //iterate
    for(kcr=kleft;kcr!=currentChr->end() && kcr->first<kright->first;kcr++){
      long kplus=getoverlaplength(kcr->first,kcr->first+windowsize, startpos,endpos);//the overlap on the right; 0 if no overlap
      long kminus=getoverlaplength(kcr->first-windowsize,kcr->first, startpos,endpos);// the overlap on the left, 0 if no overlap
      kcr->second+=(kplus-kminus)*count; // the weight on this position
    }
    
    if(startpos!=prevendpos && prevchr==chrname){
      //zero points, has no affect
    }
    prevendpos=endpos;
    prevchr=chrname;
  }
  
  ifs.close();

  //debug
//  cout<<"Scores:\n";
//  for(AllChrCvg::iterator alli=all.begin();alli!=all.end();alli++){
//    ChrCvg& crg=alli->second;
//    for(ChrCvg::iterator aci=crg.begin();aci!=crg.end();aci++){
//      cout<<alli->first<<"\t"<<aci->first<<"\t"<<aci->second<<endl;
//    }
//  }
  
  return 0;
}

/* Process candidate coverage */
int processCandidateCvg(vector<AllChrCvg>& allcvg){
  int MIN_SCORE=70*allcvg.size();
  int MIN_DIST=15;
  //cerr<<"MIN_SCORE:"<<MIN_SCORE<<endl;
  //enumerate all chromosomes of the coverage
  if(allcvg.size()==0)return -1;
  AllChrCvg::iterator ait=allcvg[0].begin();
  long prevpos=-15;
  //enumerate chromosome first
  for(;ait!=allcvg[0].end();ait++){
    string currentchr=ait->first;
    vector<ChrCvg::iterator> vitr;
    for(int i=0;i<allcvg.size();i++)
      vitr.push_back(allcvg[i][currentchr].begin());
    //enumerate pointers
    while(true){
      //check
      ChrCvg::iterator currentitr=vitr.front();
      if(currentitr==allcvg[0][currentchr].end())break;
      vector<int> scores(allcvg.size());
      for(int i=0;i<vitr.size();i++){
        if(currentitr->first!=vitr[i]->first){
          cerr<<"Error: inconsistent map contents.\n";
        }
        scores[i]=vitr[i]->second;
      }
      // output the score if it is large enough
      int sumScore=getVectorSum<int>(scores);
      if(abs(double(sumScore))>=MIN_SCORE && currentitr->first-prevpos>MIN_DIST){
        long currentpos=currentitr->first;
        cout<<currentchr<<"\t"<<currentpos<<"\t"<<currentpos<<"\t";
        if(sumScore>0)cout<<"-\t"; else cout<<"+\t";
        //type
        cout<<"C";
        for(int i=0;i<scores.size();i++) cout<<"\t"<<scores[i];
        cout<<endl;
      }
      //increase the pointer
      prevpos=currentitr->first;
      for(int i=0;i<vitr.size();i++){
        ++vitr[i];
      }
    }
  }
  
  return 0;
}

int reportZeroRegions(ChrCvg& cvg,string chr){
  ChrCvg::iterator mit=cvg.begin();
  long zerostart=-1,zeroend=-1;
  bool zerozone=false;
  while(mit!=cvg.end()){
    if(mit->second==0){
      zerostart=mit->first;
      zerozone=true;
    }else{
      if(zerozone==true){
        zeroend=mit->first;
        zerozone=false;
        processZeroRegion(zerostart, zeroend, chr);
      }
    }
    mit++;
  }
  return 0;
}
void junction2stdout_withcvgsupport(){
  AllChrJcount:: iterator jci;
  for(jci=JCOUNT.begin();jci!=JCOUNT.end();jci++){
    ChrJcount & mchi=jci->second;
    ChrJcount::iterator pit;
    for(pit=mchi.begin();pit!=mchi.end();pit++){
      vi & data=pit->second;
      int nc=0,nsample=0;
      for(int i=0;i<data.size();i++){
        nc+=data[i];
        nsample+=data[i]>0?1:0;
      }
      bool isoutput=false;
      string type="J";
      if(nc>=CMIN_JC || nsample >=CMIN_JSAMPLE){
        isoutput=true;
      }
      if(JZeroMark[jci->first].count(pit->first)>0 && JZeroMark[jci->first][pit->first]>0){
        isoutput=true;
        type+="C";
      }
      if(isoutput){
        cout<<jci->first<<"\t"<<pit->first<<"\t"<<pit->first<<"\t";
        int cdir=JDIR[jci->first][pit->first];
        if(cdir>0)  cout<<"+";
        else if(cdir<0) cout<<"-";
        else cout<<".";
        cout<<"\t"<<type;
        for(int i=0;i<data.size();i++){
          cout<<"\t"<<data[i];
        }
        cout<<endl;
      }//end if
    }//end pit for loop
  }//end jci for loop
}




int main(int argc, char* argv[]){
  if(argc<2){
    printcvgmsg();
    return 0;
  }
  //processing arguments
  int opt=1;
  vector<string> fnames;
  while(opt<argc){
    fnames.push_back(string(argv[opt]));
    opt++;
  }
  NSAMPLE=fnames.size();
  //Scan the possible junction file
  for(int i=0;i<NSAMPLE;i++){
    if(fnames[i].compare(fnames[i].length()-9,9,string(".real.wig"))==0){
      string jbedfile=fnames[i].substr(0,fnames[i].length()-9);
      jbedfile+=".junc.bed";
      cerr<<"Trying to open junction file "<<jbedfile<<"...\n";
      processonebed(jbedfile,i);
    }
  }

  //scan the coverage file and get candidate coverage 0 points
  if(NSAMPLE<1) return -1;
  for(int i=0;i<NSAMPLE;i++){
    scanCvg(fnames[i]);
  }
  
  
  vector<AllChrCvg> allcvg(NSAMPLE,JCandidateCut);
  int windowsize=100;
  for(int i=0;i<NSAMPLE;i++){
    scanCvgandHaarScore(fnames[i], allcvg[i], windowsize);
  }
  //report zero regions
 // for(int i=0;i<allcvg.size();i++){
 //   for(AllChrCvg::iterator ait=allcvg[i].begin();ait!=allcvg[i].end();ait++){
 //     ChrCvg& ccg=ait->second;
 //     reportZeroRegions(ccg,ait->first);
 //   }
 // }
  
  //report coverages
  processCandidateCvg(allcvg);
  //report junctions
  junction2stdout_withcvgsupport();
  return 0;
}
