#include <iostream>
#include <sstream>
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iomanip>
#include <glpk.h>
#include "mflow.h"
#include "menumiso.h"
#include "mpredexp.h"
#include "graphalgo.h"

/* structure for accessing parents, childrens and indices in the graph */
mi allchildren;
mi allchildrenindex;
mi allparent;
mi allparentindex;
/*the connectivity graph. allconn[i][j]=1 means i and j are connected*/
mi allconn;

/* the weight and bound for the flow problem */
vector<double> weights,bounds;


/* Parameter to stop iteration if the objective function value drops below the fraction of the maximum value */
double MAX_OBJ_FRACTION=0.01;

/* Parameter for expression level cutoff */
double EXPCUTOFF=0.5;
/* Parameter for single-exon expression level cutoff */
double SEGP0CUTOFF=0.8;
/* A modified log function */
double logu(double x){
  double LOW_BOUND=0.1;
  //if(x<LOW_BOUND)x=LOW_BOUND;
  return log(x);
}

/* Check if vector x is contained in vector y. Both x and y must be sorted and includes non-identical elements. 
   Used to check if a read is compatible with an isoform
*/
bool iscontained(vi& x, vi& y){
  if(x.size()>y.size())return false;
  
  vi::iterator nx=lower_bound(y.begin(),y.end(),x[0]);
  if(nx==y.end())return false;
  for(int i=0;i<x.size();i++,++nx){
     if(nx==y.end())return false;
    if(*nx!=x[i])return false;
  }
  return true;
}

/* Structs concerning read supporting variables and constraints */
struct S_SUPPREADS {
  /* paired-end reads accepted for constructing supporting reads */
  vi peindex;
  vi pecount;
  vi npecons; //total number of constraints for this pe type
  /* single-end reads accepted for constructing supporting reads */
  vi sgindex;
  vi sgcount;
  vi nsgcons; //total number of constraints for this sg type

  /*for each single-end read type, index of type I constraint (t_r<=f(i,i))
   Notice the type does not correspond to sgindex and sgcount
  */
  mi sgposcons; 
 /*index of type II constraint (tr<=1-f(i,i))*/
  mi sgnegcons; 

  void clear(){
    peindex.clear(); pecount.clear();
    sgindex.clear(); sgcount.clear();
    sgposcons.clear();sgnegcons.clear();
    coexpress_segs.clear();
    exclusive_segs.clear();
  }

  /* Number of variables and constraints for supporting reads*/
  int nvars;
  int nconstraints;
  
  /* Number of edges */
  int nijedges;
  /* Number of Segments.
    The total number of variables are 3*nsegs+nijedges+nvars
    Total number of constraints are 2*nsegs+nconstraints
   */
  int nsegs;

  /* The number of co-expresison and exclusive segments 
     For each segment pairs, a constraint is added.
     If x and y is coexpressed, add constraint f(x)-f(y)=0;
     if x and y is exclusive, add constraint f(x)+f(y)<=1.
  */
  mi coexpress_segs; 
  mi exclusive_segs;
  
};
S_SUPPREADS SUPPREADS;


/*
Set up the weights for supporting reads
Return the number of remaining reads 
*/
int setup_weights_4_supporting_reads(MInst& inst,  mi& res, vd& exp, double alpha);

/*
Check variables and constraints for supporting reads
Remove them to simplify the problem
*/
int remove_unused_variables(glp_prob* lp);


/* Use correlation between segments to check if they are co-expressed or exclusive? */
int setup_correlations(MInst& inst, mi& rcount,  mi& res, vd& exp, EnumParameter para);


/* Set up the weight (w_xy) and upper_bound (c_xy) of each flow f(x,y)
  In the GLPK problem, the number of variables ( edges) is 3N+ (# childrens). They're marked as:
  0..N-1: f(S, u_i)
  N..2N-1: f(v_i, T)
  2N..3N-1:f(v_i, u_i)
  the rest: f(u_i, v_j)
  res and exp is used to store existing isoforms and their expression levels of exons/isoforms already in the solution
 */
int setupflowweightandupperbound(MInst& inst, 
  mi& rcount, mi& res, vd& exp,
  EnumParameter para){

  double LAMBDA=3.0; //multiplier for the total flows
  double GAMMA=100.0;
  double POS_WEIGHT_MAX=2.0; // the upper bound for maximum positive weights
  double DEFAULT_BOUND_VAL=1.0; //the upper bound; set as 1.0
  double INTRONRETENTIONPENALTY=0.6; //penalty for intron retention. (<1)
  double SMALLNUMBER=0.1;
  double MINEXONEXP=1;

  int nsegs=inst.getNSegs();
  int nijedges=0;
  for(int i=0;i<nsegs;i++) nijedges+=allchildren[i].size();

  //record edges and exons already appear in res
  md rmap(nsegs, vd(nsegs));
  vd startmap(nsegs);
  vd endmap(nsegs);
  for(int i=0;i<res.size();i++){
    for(int j=0;j<res[i].size();j++){
      rmap[res[i][j]][res[i][j]]+=exp[i]*inst.getSegLen(res[i][j]);
    }
    for(int j=1;j<res[i].size();j++){
      rmap[res[i][j-1]][res[i][j]]+=exp[i]*(inst.getReadLen(0)-1);
    }
    if(res[i].size()>0){
      startmap[res[i][0]]+=exp[i]*inst.getSegLen(res[i][0]);
      endmap[res[i].back()]+=exp[i]*inst.getSegLen(res[i].back());
    }
  }


  weights=vector<double>(3*nsegs+nijedges,0);
  bounds=vector<double>(weights.size(),DEFAULT_BOUND_VAL);
  int index=0;
  //set up weight for f(S, u_i)
  for(int i=0;i<nsegs;i++){
    index=i;
    int nactualparents=0;
    for(int j=0;j<i;j++) 
      nactualparents+=rcount[j][i];
    if(nactualparents==0){
    //if(allparent[i].size()==0){
      weights[index]=LAMBDA*1.0; // must be positive weight
    }else{
      weights[index]=LAMBDA*(1.0-GAMMA); //must be negative weight to prevent "short-cuts"
    }
    
  }
  //set up weight for f(v_i,T)
  for(int i=0;i<nsegs;i++){
    index=i+nsegs;
    int nactualchildren=0;
    for(int j=i+1;j<nsegs;j++)
      nactualchildren+=rcount[i][j];
    //nactualparents=allchildren[i].size();
    if(nactualchildren==0 ){
      weights[index]=0.0;
    }else{
      weights[index]=(0.0-GAMMA); //must be negative weight to prevent "short-cuts"
    }
  }
  //set up weight for f(u_i,v_i)
  //vector<int> exonexpi=getRowSum<int>(inst.segcount);
  //vector<double>exonexp(exonexpi.size());
  //for(int i=0;i<exonexp.size();i++) exonexp[i]=exonexpi[i]/inst.getSegLen(i);
  vector<double> exonexp;
  if(!para.assemble_by_sample)
    exonexp=getRowSum<double>(inst.segcvg);
  else{
    exonexp.resize(inst.segcvg.size(),0);
    for(int i=0;i<inst.segcvg.size();i++) exonexp[i]=inst.segcvg[i][para.current_sample];
  }
  for(int i=0;i<exonexp.size();i++)exonexp[i]/=inst.getNSamples();
  for(int i=0;i<nsegs;i++){
    index=i+nsegs*2;
    double x=exonexp[i]-rmap[i][i];
    cout<<"Segment "<<i<<",exonexp="<<x<<endl;
    if(x<MINEXONEXP)x=MINEXONEXP;
    //use segp0, get the min segp0
    double minsegp0=*min_element(inst.segp0[i].begin(),inst.segp0[i].end());
    weights[index]=log(x+1);
    //weights[index]*=(1-minsegp0)*(1-minsegp0); //must be positive weights
  }
  //set up weight for f(u_i,v_j)
  // calculate log probability for summing v_j 
  for(int i=0;i<nsegs;i++){
    //calculate the fractions of: junctions and exon expressions
    double totalj=0.0;
    double totalexp=0.0;
    vd vjtotal, vexptotal;
    //first, sum up the total number of junction counts and exon coverages
    for(int j=0;j<allchildren[i].size();j++){
      int pd=allchildren[i][j];
      double x1=rcount[i][pd]-rmap[i][pd];
      double x2=exonexp[pd]-rmap[pd][pd];
      if(x1<SMALLNUMBER)x1=SMALLNUMBER;
      if(x2<SMALLNUMBER)x2=SMALLNUMBER;
      if(pd==i+1){
        x1*=INTRONRETENTIONPENALTY;
        x2*=INTRONRETENTIONPENALTY;
      }
      vjtotal.push_back(x1);
      vexptotal.push_back(x2);
    }
    for(int j=0;j<allchildren[i].size();j++){
      totalj+=vjtotal[j];
      totalexp+=vexptotal[j];
    }
    if(totalj==0)totalj=1;
    if(totalexp==0)totalexp=1;
    for(int j=0;j<allchildren[i].size();j++){
      index=allchildrenindex[i][j];
      double logpij=logu(vjtotal[j]/totalj);
      double logpe=logu(vexptotal[j]/totalexp);
      weights[index]=logpij+logpe; //negative 
    }
  }
  // calculate log probability for summing u_i 
  for(int i=0;i<nsegs;i++){
    //calculate the fractions of: junctions and exon expressions
    double totalj=0.0;
    double totalexp=0.0;
    vd vjtotal, vexptotal;
    //first, sum up the total number of junction counts and exon coverages
    //cout<<"Sum u_i:node "<<i<<":\n";
    for(int j=0;j<allparent[i].size();j++){
      int pd=allparent[i][j];
      double x1=rcount[pd][i]-rmap[pd][i];
      double x2=exonexp[pd]-rmap[pd][pd];
      if(x1<SMALLNUMBER)x1=SMALLNUMBER;
      if(x2<SMALLNUMBER)x2=SMALLNUMBER;
      if(pd==i+1){
        x1*=INTRONRETENTIONPENALTY;
        x2*=INTRONRETENTIONPENALTY;
      }
      vjtotal.push_back(x1);
      vexptotal.push_back(x2);
    }
    for(int j=0;j<allparent[i].size();j++){
      totalj+=vjtotal[j];
      totalexp+=vexptotal[j];
    }
    if(totalj==0)totalj=1;
    if(totalexp==0)totalexp=1;
    for(int j=0;j<allparent[i].size();j++){
      double logpij=logu(vjtotal[j]/totalj);
      double logpe=logu(vexptotal[j]/totalexp);
      index=allparentindex[i][j];
      weights[index]+=logpij+logpe; // negative
      //cout<<allparent[i][j]<<",("<<logpij<<")("<<logpe<<")index("<<index<<"),w("<<weights[index]<<") ";
    }
    //cout<<endl;
  }

  return 0;
}

/* Do not override existing weights and bounds, but instead update them */
int renewweightandbound(MInst& inst, mi& rcount, mi&res, vd& exp){
  double ADJUSTWEIGHT=0.9;
  double ADDITIVEWEIGHT=-0.5;
  double DECAY=0.9; //even if the res keep unchanged, decade the weights if necessary
  double BRANCHPENALTY=0.6;//for branches (segment connects to more than 1 segment), enforce higher penalty
  double INTRONRETENTIONPENALTY=0.6;
  int nsegs=inst.getNSegs();

  //record edges and exons already appear in res
  md rmap(nsegs, vd(nsegs));
  vd startmap(nsegs);
  vd endmap(nsegs);
  for(int i=0;i<res.size();i++){
    for(int j=0;j<res[i].size();j++){
      rmap[res[i][j]][res[i][j]]+=(exp[i])*inst.getSegLen(res[i][j])+1e-3;
    }
    for(int j=1;j<res[i].size();j++){
      rmap[res[i][j-1]][res[i][j]]+=(exp[i])*exp[i]*(inst.getReadLen(0)-1)+1e-3;
    }
    if(res[i].size()>0){
      startmap[res[i][0]]+=exp[i]*inst.getSegLen(res[i][0])+1e-3;
      endmap[res[i].back()]+=exp[i]*inst.getSegLen(res[i].back())+1e-3;
    }
  }

  //set up weight for f(S, u_i)
  int index;
  for(int i=0;i<nsegs;i++){
    index=i;
    if(startmap[i]>0){
      weights[index]+=ADDITIVEWEIGHT;
    }
  }
  //set up weight for f(v_i,T)
  for(int i=0;i<nsegs;i++){
    index=i+nsegs;
    if(endmap[i]>0){
      weights[index]+=ADDITIVEWEIGHT;
    }
  }
  //vector<double> exonexp=getRowSum<double>(inst.segcvg);
  for(int i=0;i<nsegs;i++){
    index=i+nsegs*2;
    if(rmap[i][i]>0){
      weights[index]*=ADJUSTWEIGHT; //positive weight, reduce
    }else{
      //weights[index]/=ADJUSTWEIGHT; //positive weight, increase 
    }
  }
  // calculate log probability for summing v_j 
  for(int i=0;i<nsegs;i++){
    bool branch=false;
    if(allchildren[i].size()>1)branch=true;
    for(int j=0;j<allchildren[i].size();j++){
      index=allchildrenindex[i][j];
      int c=allchildren[i][j];
      if(rmap[i][c]>0){
          weights[index]/=ADJUSTWEIGHT; //negative, increase
        if(branch){
          //provide even higher penalty for branches
          weights[index]/=BRANCHPENALTY;
        }
        if(c==i+1){
          weights[index]/=INTRONRETENTIONPENALTY;
        }
        //if(weights[index]>-1e-5)weights[index]=-100; // permanently disable this connection
      }else{
          //weights[index]*=ADJUSTWEIGHT; //negative, increase
        if(branch){
          //provide even higher penalty for branches
          //weights[index]*=BRANCHPENALTY;
        }
        if(c==i+1){
          //weights[index]*=INTRONRETENTIONPENALTY;
        }
      }
    }
  }


  for(int i=0;i<weights.size();i++){
    if(weights[i]>=0)weights[i]*=DECAY;
    else weights[i]/=DECAY;
  }
  return 0;
}

/* set up all matrix elements to glp_prob */
int setup_all_mats(glp_prob* lp,MInst& inst);

int updateboundandweight(MInst& inst, glp_prob* lp,mi& rcount, mi& res, vd& exp,EnumParameter para){
  int nsegs=inst.getNSegs();

  // for each flow, set up its upper bound and weights
  if(weights.size()==0)
    setupflowweightandupperbound(inst,  rcount, res,exp,para);
  else
    renewweightandbound(inst,rcount,res,exp);

  //set up the upper bound for supporting reads
  setup_weights_4_supporting_reads(inst,   res,  exp, para.RDSUPPORT_ALPHA);


  //remove_unused_variables(lp);
  setup_all_mats(lp,inst);

  
  for(int index=0;index<weights.size();index++){
    //glp_set_col_bnds(lp,index+1,GLP_DB, 0.0, 1.0); // set up 0<=f(S,u_i)<=C here
    glp_set_obj_coef(lp,index+1,weights[index]); //set up w_ij for all flows here
  }
  return 0;
}


double getentropy(vd x){
  double sumx=0;
  for(int i=0;i<x.size();i++) sumx+=x[i];
  //normalize
  if(sumx==0) x=vd(x.size(),1.0/x.size());
  else 
    for(int i=0;i<x.size();i++) x[i]/=sumx;
  double entropy=0;
  for(int i=0;i<x.size();i++) {
    if(x[i]>0) entropy-=x[i]*log(x[i]);
  }
  return entropy;
}

/* Get the jenson-shannon distribution */
double getjensonshannon(vd x, vd y){
  double ds=0;
  if(x.size()!=y.size()){
    cerr<<"Error: to calculate Jensen-Shannon divergence, both vector lengths must be equal.\n";
    return 0;
  }
  //normalize
  double sumx=0,sumy=0;
  for(int i=0;i<x.size();i++){
    sumx+=x[i];
    sumy+=y[i];
  }
  if(sumx==0) x=vd(x.size(),(1.0/(int)x.size()));
  else for(int i=0;i<x.size();i++) x[i]/=sumx;
  if(sumy==0) y=vd(y.size(),(1.0/(int)y.size()));
  else for(int i=0;i<y.size();i++) y[i]/=sumy;
  vd xplusy(x.size(),0);
  for(int i=0;i<xplusy.size();i++) xplusy[i]=(x[i]+y[i])/2;
  double exy=getentropy(xplusy);
  double ex=getentropy(x);
  double ey=getentropy(y);
  return exy-(ex+ey)/2;
}

/* Use correlation between segments to check if they are co-expressed or exclusive? */
int setup_correlations(MInst& inst,  mi& rcount,  EnumParameter para){
  int nsamples=inst.getNSamples();
  if(nsamples<2) return -1;
  int nsegs=inst.getNSegs();
  int rdlen=inst.getReadLen(0);
  //choose segments
  vi chosensegments;
  vi chosenmap(nsegs,0);
  for(int i=0;i<nsegs;i++){
    if(inst.getSegLen(i)>=0) {
      chosensegments.push_back(i);
      chosenmap[i]=1;
    }
  }
  cout<<"Chosen segments:\n"; for(int i=0;i<chosensegments.size();i++)cout<<chosensegments[i]<<" "; cout<<endl;

  //calculate the pariwise jenson-shannon distance 
  md jsm(chosensegments.size(),vd(chosensegments.size(),0.0));
  vd jscollect;
  vd samplerdcnt(nsamples,0);
  for(int i=0;i<nsamples;i++) {samplerdcnt[i]=inst.getSampleRdCnt(i); if(samplerdcnt[i]==0) samplerdcnt[i]=1; }
  for(int i=0;i<jsm.size();i++){
    int ci=chosensegments[i];
    vd nci=inst.segcvg[ci];
    //normalize by total read count of the gene
    for(int ii=0;ii<nci.size();ii++) nci[ii]/=samplerdcnt[ii];

    for(int j=i+1;j<jsm.size();j++){
      int cj=chosensegments[j];
      vd ncj=inst.segcvg[cj];
      for(int jj=0;jj<ncj.size();jj++) ncj[jj]/=samplerdcnt[jj];

      double jsval=getjensonshannon(nci,ncj);
      double jsd=sqrt(jsval);
      //cout<<"JS["<<ci<<"]["<<cj<<"]="<<jsd<<endl;
      jsm[i][j]=jsd;
      //collect into js
      if(j!=i+1 && inst.getSegLen(i)>=para.JS_MIN_SEG_LEN_TO_CONSIDER 
       && inst.getSegLen(j)>=para.JS_MIN_SEG_LEN_TO_CONSIDER){
        jscollect.push_back(jsd);
      }
    }
  }
  cout<<"Summary of JS:\n";
  for(int i=0;i<jscollect.size();i++){
    if(i>0)cout<<",";
    cout<<jscollect[i];
  }
  cout<<endl;
  double alljsmean=getmean(jscollect);
  double alljsstd=getstd(jscollect);
  cout<<"Mean: "<<alljsmean<<",std:"<<alljsstd<<endl;

  //reference segments
  mi refsegs;
  vd jsref;
  getreferencesegments(inst, rcount,allchildren, allparent, para, refsegs);
  cout<<"Ref segments:\n";
  for(int i=0;i<refsegs.size();i++){
    for(int j=0;j<refsegs[i].size();j++) cout<<refsegs[i][j]<<" ";
    cout<<endl;
  }
  for(int i=0;i<refsegs.size();i++){
    for(int j=0;j<refsegs[i].size();j++){
      int seg1=refsegs[i][j];
      for(int k=j+1;k<refsegs[i].size();k++){
        int seg2=refsegs[i][k];
        //if(j!=0 || k!=j+1) cout<<","; cout<<jsm[seg1][seg2];
        jsref.push_back(jsm[seg1][seg2]);
      }
    }
    cout<<endl;
  }
  cout<<"Ref JS total:\n";
  for(int i=0;i<jsref.size();i++){
    if(i>0)cout<<",";
    cout<<jsref[i];
  }
  cout<<endl;
  sort(jsref.begin(),jsref.end());
  if(jsref.size()>10) jsref=vd(jsref.begin(),jsref.begin()+(int)jsref.size()*0.8);
  double refjsmean=getmean(jsref);
  double refjsstd=getstd(jsref);
  cout<<"Mean: "<<refjsmean<<",std:"<<refjsstd<<endl;


  //check ref JS or total JS are differentialble?
  bool validjsdist=false;
  if(alljsmean>refjsmean+para.JS_MIN_DISTRIB_DIST*refjsstd) {
    validjsdist=true;
    cout<<"Two distributions are differentiable.\n";
  }else{
    cout<<"Two distributions are NOT differentiable.\n";
    return 0;
  }



  //determine segments to check
  mi segstocheck;
  getcheckingsegments(inst, rcount,allchildren, allparent, para, segstocheck);
  cout<<"Checking segments:\n";
  int MIN_JUNC_AS_GOOD=10;
  for(int i=0;i<segstocheck.size();i++){
    vi apre=adjustparent(allparent, segstocheck[i][0],inst,para,rcount);
    vi ache=adjustchildren(allchildren,segstocheck[i][1],inst,para,rcount);
    int nvalidparents=0;
    int nvalidchildren=0;
    
    cout<<"parents of "<<segstocheck[i][0]<<" ("<<apre.size()<<", ";
    for(int j=0;j<apre.size();j++) {
      int nj=rcount[apre[j]][segstocheck[i][0]];
      cout<<apre[j]<<":"<<nj<<" ";
      if(nj>=MIN_JUNC_AS_GOOD) nvalidparents++;
    }
    cout<<")";
    cout<<" vs children of "<<segstocheck[i][1]<<" ("<<ache.size()<<", ";
    for(int j=0;j<ache.size();j++){
      int nj=rcount[segstocheck[i][1]][ache[j]];
      cout<<ache[j]<<":"<<nj<<" ";
      if(nj>=MIN_JUNC_AS_GOOD) nvalidchildren++;
    }
    cout<<")"<<endl;
    cout<<"parents of "<<segstocheck[i][0]<<": "<<inst.chrname<<":"<<inst.segment[segstocheck[i][0]][0]<<endl;
    mi parentcluster,childrencluster;
    getparentchain(inst, allchildren, allparent, segstocheck[i][0], parentcluster,para,rcount );
    for(int j=0;j<parentcluster.size();j++){
      cout<<"\t";
      for(int k=0;k<parentcluster[j].size();k++)
        cout<<parentcluster[j][k]<<" ";
      cout<<inst.chrname<<":"<<inst.segment[parentcluster[j][0]][1];
      cout<<endl;
    }

    cout<<"children of "<<segstocheck[i][1]<<": "<<inst.chrname<<":"<<inst.segment[segstocheck[i][1]][1]<<endl;
    getchildrenchain(inst, allchildren, allparent, segstocheck[i][1], childrencluster,para,rcount );
    for(int j=0;j<childrencluster.size();j++){
      cout<<"\t";
      for(int k=0;k<childrencluster[j].size();k++)
        cout<<childrencluster[j][k]<<" ";
      cout<<inst.chrname<<":"<<inst.segment[childrencluster[j][0]][0];
      cout<<endl;
    }
    if(parentcluster.size()==0 || childrencluster.size()==0 || nvalidparents<2 || nvalidchildren<2) continue;
    cout<<"pairwise comparisons:\n";
    for(int j1=0;j1<parentcluster.size();j1++){
      for(int j2=0;j2<childrencluster.size();j2++){
        //perform all-vs-all pairwise checking 
        if(parentcluster[j1].size()==0 || childrencluster[j2].size()==0) continue;
        vd checkjsms;
        cout<<"Checking group "<<j1<<" (p) vs group "<<j2<<" (c)\n";
        for(int k1=0;k1<parentcluster[j1].size();k1++){
          for(int k2=0;k2<childrencluster[j2].size();k2++){
            double crjsm=jsm[parentcluster[j1][k1]][childrencluster[j2][k2]];
            cout<<"("<<parentcluster[j1][k1]<<","<<childrencluster[j2][k2]<<")=";
            cout<<crjsm<<",";
            checkjsms.push_back(crjsm);
          }
        }
        cout<<endl;
        //perform test
        double cjsmmean=getmean(checkjsms);
        double cjsmstd=getstd(checkjsms);
        if(cjsmmean>alljsmean+para.JS_MIN_EX_DIST*alljsstd){
           vi coex; coex.push_back(parentcluster[j1][0]);coex.push_back(childrencluster[j2][0]);
           bool isalready=false;
           for(int sii=0;sii<SUPPREADS.exclusive_segs.size();sii++) if(SUPPREADS.exclusive_segs[sii]==coex) isalready=true;
           if(!isalready){
             SUPPREADS.exclusive_segs.push_back(coex);
             cout<<"Instance:"<<inst.id<<" JS:Possible exclusion exons; ";
             cout<<"Adding "<<coex[0]<<" and "<<coex[1]<<endl;
             if(inst.refs.size()>0){
               //checking annotation
               if(isexclusiveinanno(inst, coex[0], coex[1])==1)  cout<<"Inst: "<<inst.id<<" CHKREF: confirmed.\n";
             }
           }
        }
        if(cjsmmean<refjsmean+para.JS_MIN_CO_DIST*refjsstd){
           vi coex; coex.push_back(parentcluster[j1][0]);coex.push_back(childrencluster[j2][0]);
           bool isalready=false;
           for(int sii=0;sii<SUPPREADS.coexpress_segs.size();sii++) if(SUPPREADS.coexpress_segs[sii]==coex) isalready=true;
           if(!isalready){
             SUPPREADS.coexpress_segs.push_back(coex);
             cout<<"Instance:"<<inst.id<<" JS:Possible coexpression exons; ";
             cout<<"Adding "<<coex[0]<<" and "<<coex[1]<<endl;
             if(inst.refs.size()>0){
               //checking annotation
               if(iscoexpressioninanno(inst, coex[0], coex[1])==1)  cout<<"Inst: "<<inst.id<<" CHKREF: confirmed.\n";
             }
           }
        }
      }
    }
    cout<<endl;
  }

  return 0;
}

/* Get the supporting read information */
int getsupportingreadinfo(MInst& inst,EnumParameter para){

  
  int nsegs=inst.getNSegs();
  cout<<"Instance "<<inst.id<<", nsegs:"<<nsegs<<endl;
  //for each nodes of the instance, select the number of its childrens.
  //This will determine the number of columns.
  int nijedges=0;
  for(int i=0;i<nsegs;i++){
    nijedges+=allchildren[i].size();
  }
  SUPPREADS.clear();
  SUPPREADS.nsegs=nsegs;
  SUPPREADS.nijedges=nijedges;

  //count the number of variables and constraints for single-end reads
  vi nsgcons;
  vi nsgreads;
  for(int i=0;i<inst.type.size();i++){
    int si=inst.type[i].front();
    int sj=inst.type[i].back();
    nsgcons.push_back(2*inst.type[i].size()-1); //this is the number of constraints to be added for single-end
    if(!para.assemble_by_sample)
      nsgreads.push_back(getVectorSum<int>(inst.typecount[i]));
    else
      nsgreads.push_back(inst.typecount[i][para.current_sample]);
    //fill sgposcons and sgnegcons: which segment is included into the constraint?
    SUPPREADS.sgposcons.push_back(vi());
    SUPPREADS.sgnegcons.push_back(vi());
    int typeind=0;
    for(int j=si;j<=sj;j++){
      if(j==inst.type[i][typeind]){
        //constraints for pos inclusion; i.e., t_r<=f(i,i)
        SUPPREADS.sgposcons[i].push_back(j);
        typeind++;
      }else{
        //constraints for neg inclusion; i.e., t_r<=1-f(i,i)
        SUPPREADS.sgnegcons[i].push_back(j);
      }
    }
  }
  //check if single-end types are valid
  vi isvalid(inst.type.size(),1);
  for(int i=0;i<inst.type.size();i++){
    //check if the segments are connected or not
    for(int j=0;j<inst.type[i].size()-1;j++){
      int e1=inst.type[i][j];
      int e2=inst.type[i][j+1];
      bool haschildren=false;
      for(int k=0;k<allchildren[e1].size();k++){
        if(allchildren[e1][k]==e2){
          haschildren=true;
          break;
        }
      }
      if(haschildren==false){
        isvalid[i]=0;
        break;
      }
    }
  }  
  //check paired-end reads
  vi npecons,npevars;
  vi npereads;
  int nvars=0,nconstraints=0;

  SUPPREADS.nsgcons.clear();
  SUPPREADS.npecons.clear();
  for(int i=0;i<inst.petype.size();i++){
    int pi=inst.petype[i].front();
    int pj=inst.petype[i].back();
    //substract paired-end read counts from single-end read counts
    int allpcount;
    if(!para.assemble_by_sample)
      allpcount==getVectorSum<int>(inst.petypecount[i]);
    else
      allpcount=inst.petypecount[i][para.current_sample];
    nsgreads[pi]-=allpcount;
    nsgreads[pj]-=allpcount;
    //both types of single-end reads must be valid
    if(isvalid[pi]==0 || isvalid[pj]==0)continue;
    int pilast=inst.type[pi].back();
    int pjfirst=inst.type[pj].front();
    //both pairs must be connected
    if(allconn[pilast][pjfirst]==0)continue;

    npecons.push_back(nsgcons[pi]); //add #constraints of 1
    if(pj!=pi) npecons.back()+=nsgcons[pj];
    // update variables and constraints
    nvars++;
    nconstraints+=npecons.back();
    cout<<"PE 1:"<<pi<<"("<<nsgcons[pi]<<"),PE2:"<<pj<<"("<<nsgcons[pj]<<")\n";
    cout<<"PE:"<<i<<",NCONSTRAINTS:"<<npecons.back()<<",total:"<<nconstraints<<endl;


    SUPPREADS.peindex.push_back(i);
    SUPPREADS.pecount.push_back(allpcount);
    SUPPREADS.npecons.push_back(npecons.back());
  }
  //add remaining single-end reads
  for(int i=0;i<inst.type.size();i++){
    if(nsgreads[i]>0 && isvalid[i]>0){
      SUPPREADS.sgindex.push_back(i);
      SUPPREADS.sgcount.push_back(nsgreads[i]);
      nvars++;
      nconstraints+=nsgcons[i];
      cout<<"SG:"<<i<<",NCONSTRAINTS:"<<nsgcons[i]<<",total:"<<nconstraints<<endl;
      SUPPREADS.nsgcons.push_back(nsgcons[i]);
    }
  }
  SUPPREADS.nvars=nvars;
  SUPPREADS.nconstraints=nconstraints;

  if(nconstraints>=para.MAX_SUPP_CONSTRAINT){
    SUPPREADS.nvars=0;
    SUPPREADS.nconstraints=0;
    SUPPREADS.clear();
  }
  
  
  return 0;
}

/* Update SUPPREADS 
remove records with 0 entries
*/
int update_supporting_readinfo(MInst& inst){
  //re-calculate nvars and nconstraints
  int nvars=0;
  int nconstraints=0;
  
  vi peindex,pecount;
  vi sgindex,sgcount;
  vi npecons,nsgcons;

  for(int i=0;i<SUPPREADS.peindex.size();i++){
    if(SUPPREADS.pecount[i]==0)continue;
    nvars++;
    int petype=SUPPREADS.peindex[i];
    nconstraints+=SUPPREADS.npecons[i];
    peindex.push_back(petype);
    pecount.push_back(SUPPREADS.pecount[i]);
    npecons.push_back(SUPPREADS.npecons[i]);
  }

  for(int i=0;i<SUPPREADS.sgindex.size();i++){
    if(SUPPREADS.sgcount[i]==0)continue;
    nvars++;
    int sgtype=SUPPREADS.sgindex[i];
    nconstraints+=SUPPREADS.nsgcons[i];
    sgindex.push_back(sgtype);
    sgcount.push_back(SUPPREADS.sgcount[i]);
    nsgcons.push_back(SUPPREADS.nsgcons[i]);
  }
  SUPPREADS.nvars=nvars;
  SUPPREADS.nconstraints=nconstraints;

  //replace the original content
  SUPPREADS.peindex=peindex;
  SUPPREADS.pecount=pecount;
  SUPPREADS.sgindex=sgindex;
  SUPPREADS.sgcount=sgcount;
  SUPPREADS.npecons=npecons;
  SUPPREADS.nsgcons=nsgcons;
 
  cout<<"NSEGS:"<<SUPPREADS.nsegs<<endl;
  cout<<"NEDGES:"<<SUPPREADS.nijedges<<endl;
  cout<<"NVARS:"<<SUPPREADS.nvars<<endl;
  cout<<"NCONSTRAINTS:"<<SUPPREADS.nconstraints<<endl;


  return 0;
}
/* The following variables are used to set up GLPK matrix coefficients*/
int MAT_ARRAY_IND;
int* MAT_RIND;
int* MAT_CIND;
double* MAT_VIND;
int MAT_ARRAY_LEN=0;
/* Functions to reset the above variables
vlen passes the array size into the function.
If vlen=-1, delete the arrays */
void reset_mat_element(int vlen){
  MAT_ARRAY_IND=1;
  if(vlen>MAT_ARRAY_LEN || vlen==-1){
    if(MAT_ARRAY_LEN>0){
      delete []MAT_RIND;
      delete []MAT_CIND;
      delete []MAT_VIND;
    }
    MAT_RIND=new int[vlen];
    MAT_CIND=new int[vlen];
    MAT_VIND=new double[vlen];
    MAT_ARRAY_LEN=vlen;
  }
  for(int i=0;i<MAT_ARRAY_LEN;i++){
    MAT_RIND[i]=0;
    MAT_CIND[i]=0;
    MAT_VIND[i]=0;
  }
}
/*
If the MAT_RIND is too small, dynamically enlarge it and copy the matrix
len must satisfy len>MAT_ARRAY_LEN
*/
void enlarge_mat_element(int len){
  cout<<"Enlarging array from "<<MAT_ARRAY_LEN<<" to "<<len<<endl;
  int* p_mr=MAT_RIND;
  int* p_mc=MAT_CIND;
  double* p_mv=MAT_VIND;
  
  //resize
  MAT_RIND=new int[len];
  MAT_CIND=new int[len];
  MAT_VIND=new double[len];
  for(int i=0;i<len;i++){
    MAT_RIND[i]=0; MAT_CIND[i]=0; MAT_VIND[i]=0;
  }
  //copy
  for(int i=0;i<MAT_ARRAY_LEN;i++){
    MAT_RIND[i]=p_mr[i];
    MAT_CIND[i]=p_mc[i];
    MAT_VIND[i]=p_mv[i];
  }
  MAT_ARRAY_LEN=len;
  //delete
  delete []p_mr;
  delete []p_mc;
  delete []p_mv;
}
/* Functions to modify matrix contents */
void set_mat_element(int r, int c, double v){
  if(MAT_ARRAY_IND>=MAT_ARRAY_LEN) enlarge_mat_element(MAT_ARRAY_LEN*2);
  MAT_RIND[MAT_ARRAY_IND]=r;
  MAT_CIND[MAT_ARRAY_IND]=c;
  MAT_VIND[MAT_ARRAY_IND]=v;
  //cout<<"SETTING MAT["<<r<<","<<c<<"]="<<v<<",index="<<MAT_ARRAY_IND<<endl;
  MAT_ARRAY_IND++;
}
/* set up constraints for single-end read type, at column colindex 
 Parameters: 
 colindex: the index of the column (constraints).
 nv: read supporting variable index (rows)
 pi: the index of sgtypes to add
*/
void set_sgtype_const(glp_prob* lp,int &colindex,int nv, int pi){
    int nsegs=SUPPREADS.nsegs;
    
    for(int nps=0;nps<SUPPREADS.sgposcons[pi].size();nps++){
      int ccons=SUPPREADS.sgposcons[pi][nps];
      //set up constraints
      //for variable t_r
      set_mat_element(colindex,nv,1.0);
      //for variable f(i,i)
      set_mat_element(colindex,2*nsegs+ccons+1,-1.0);
      glp_set_row_bnds(lp,colindex,GLP_UP,0.0,0.0);
      colindex++;
    }
    for(int nps=0;nps<SUPPREADS.sgposcons[pi].size()-1;nps++){
      int ccons=SUPPREADS.sgposcons[pi][nps];
      int cconschildren=SUPPREADS.sgposcons[pi][nps+1];
      int ctindex=-1;
      //figure out index of the edges
      for(int j=0;j<allchildren[ccons].size();j++){
        if(allchildren[ccons][j]==cconschildren){
          ctindex=allchildrenindex[ccons][j];
        }
      }
      if(ctindex==-1){
        cerr<<"Error: cannot find index.\n";
      }
      //set up constraints
      //for variable t_r
      set_mat_element(colindex,nv,1.0);
      //for variable f(i,j)
      set_mat_element(colindex,ctindex+1,-1.0);
      glp_set_row_bnds(lp,colindex,GLP_UP,0.0,0.0);
      colindex++;
    }
    //for(int nps=0;nps<SUPPREADS.sgnegcons[pi].size();nps++){
    //  int ccons=SUPPREADS.sgnegcons[pi][nps];
    //  //set up constraints
    //  //for variable t_r
    //  set_mat_element(colindex,nv,1.0);
    //  //for variable f(i,i)
    //  set_mat_element(colindex,2*nsegs+ccons+1,1.0);
    //  glp_set_row_bnds(lp,colindex,GLP_UP,0.0,1.0);
    //  colindex++;
    //}

}

/* Set up the constraints for coexpression or exclusive segments */
void set_coexclu_constraints(glp_prob* lp,int &colindex){
  for(int i=0;i<SUPPREADS.coexpress_segs.size();i++){
    cout<<"Setting constraints for coexpressed segments "<<SUPPREADS.coexpress_segs[i][0]<<" and "<<SUPPREADS.coexpress_segs[i][1]<<", index="<<colindex<<endl;
    set_mat_element(colindex, SUPPREADS.coexpress_segs[i][0]+1, 1.0);
    set_mat_element(colindex, SUPPREADS.coexpress_segs[i][1]+1, -1.0);
    glp_set_row_bnds(lp,colindex,GLP_FX,0.0,0.0); // f(x)-f(y)=0
    colindex++;
  }
  for(int i=0;i<SUPPREADS.exclusive_segs.size();i++){
    cout<<"Setting constraints for exclusive segments "<<SUPPREADS.exclusive_segs[i][0]<<" and "<<SUPPREADS.exclusive_segs[i][1]<<", index="<<colindex<<endl;
    set_mat_element(colindex, SUPPREADS.exclusive_segs[i][0]+1, 1.0);
    set_mat_element(colindex, SUPPREADS.exclusive_segs[i][1]+1, 1.0);
    glp_set_row_bnds(lp,colindex,GLP_DB,0.0,1.0); // f(x)-f(y)=0
    colindex++;
  }
}

int setup_all_mats(glp_prob* lp,MInst& inst){

  int nsegs=SUPPREADS.nsegs;
  int nijedges=SUPPREADS.nijedges;
  //set up constraint matrix
  int vnlen=4*nsegs+2*SUPPREADS.nijedges+2*SUPPREADS.nconstraints+1000;
  //arrays to set up matrix
  cout<<"vnlen:"<<vnlen<<endl;
  reset_mat_element(vnlen);

  //reset lp's number of rows and columns
  //Number of variables include s->u, u->v, v->t (3*nsegs), junction edges (nijedges) and read supprting variables (SUPPREADS.nvars)
  int sncols=3*nsegs+nijedges+SUPPREADS.nvars;//variables
  //Number of constraints include flow constraints for each u and v (2*nsegs), read supporting constraints (SUPPREADS.nconstraints),
  // 1 constraint of the sum of the flow (1), and all co-express and exclusive express segments.
  int snrows=2*nsegs+SUPPREADS.nconstraints+1+SUPPREADS.coexpress_segs.size()+SUPPREADS.exclusive_segs.size();//constraints

  //number of variables
  glp_add_cols(lp,sncols);
  //number of constraints
  glp_add_rows(lp,snrows);
 
  /*
  int currentcols=glp_get_num_cols(lp);
  int currentrows=glp_get_num_rows(lp);
  //delete columns
  if(sncols<currentcols){
    int todel=currentcols-sncols;
    int* num=new int[todel+1];
    for(int i=0;i<todel;i++) num[i+1]=sncols+i+1;
    glp_del_cols(lp,todel,num);
    delete []num;
  }
  //delete rows
  if(snrows<currentrows){
    int todel=currentrows-snrows;
    int* num=new int[todel+1];
    for(int i=0;i<todel;i++) num[i+1]=snrows+i+1;
    glp_del_rows(lp,todel,num);
    delete []num;
  }
  */
  
  //variable bounds, between 0-1
  for(int i=0;i<3*nsegs+nijedges+SUPPREADS.nvars;i++)
    glp_set_col_bnds(lp,i+1,GLP_DB,0.0,1.0);
    
  //flow constraints, index from 1..2*nsegs, bound=0
  for(int i=0;i<2*nsegs;i++)
    glp_set_row_bnds(lp,i+1,GLP_FX,0.0,0.0);

 
  //add constraint matrix rows (still flow constraints)
  for(int i=0;i<nsegs;i++){
    // from S, u_i
    set_mat_element(i+1,i+1,1.0);
    //from u_i to v_i
    set_mat_element(i+1,2*nsegs+i+1,-1.0);
    //from v_i to T
    set_mat_element(nsegs+i+1,nsegs+i+1,-1.0);
    //from u_i to v_i
    set_mat_element(nsegs+i+1,2*nsegs+i+1,1.0);
  }
  //for all edges starting from v_i
  for(int i=0;i<nsegs;i++){
    for(int j=0;j<allparentindex[i].size();j++){
      int pindex=allparentindex[i][j];
      set_mat_element(i+1,pindex+1,1.0);
    }
  }
  for(int i=0;i<nsegs;i++){
    for(int j=0;j<allchildrenindex[i].size();j++){
      int pindex=allchildrenindex[i][j]; 
      set_mat_element(nsegs+i+1,pindex+1,-1.0);
    }
  }

  //set up constraints for read support variables (t_r)
  int colindex=2*nsegs+1;
  int varindex=3*nsegs+nijedges+1;
  for(int npe=0;npe<SUPPREADS.peindex.size();npe++){
    int nv=varindex;//the index of the variable
    varindex++;
    int pei=SUPPREADS.peindex[npe];//the corresponding paired-end index
    int pi=inst.petype[pei].front();//the corresponding two single-end index
    int pj=inst.petype[pei].back();
    //set up the constraints for the 1st pair
    //if the second pair is different from the first pair,
    int colindex0=colindex;
    set_sgtype_const(lp,colindex,nv,pi);
    //set up the constraints for the 2nd pair
    if(pj!=pi){
      set_sgtype_const(lp,colindex,nv,pj);
    }
    cout<<"Setting constraints for PE "<<pei<<",row: "<<colindex0<<"-"<<colindex-1<<endl;
  }
  //set up constraints for single-end reads
  for(int nsg=0;nsg<SUPPREADS.sgindex.size();nsg++){
    int nv=varindex;//the index of the variable
    varindex++;
    int sei=SUPPREADS.sgindex[nsg];
    int colindex0=colindex;
    set_sgtype_const(lp,colindex,nv,sei);
    cout<<"Setting constraints for SG "<<sei<<",row: "<<colindex0<<"-"<<colindex-1<<endl;
  }
 
  
  //set up co-expression and exclusive segments
  set_coexclu_constraints( lp,colindex);
  
  //set up the flow constraint, make sure that the sum of the flow is 1
  glp_set_row_bnds(lp,colindex,GLP_DB,0.0,1.0);
  for(int i=0;i<nsegs;i++)
    set_mat_element(colindex,i+1,1.0);
  colindex++;
  
  //load the matrix into the GLP problem
  glp_load_matrix(lp,MAT_ARRAY_IND-1,MAT_RIND,MAT_CIND,MAT_VIND);
  
}


/*
Set up the weights for supporting reads
Return the number of remaining reads 
*/
int setup_weights_4_supporting_reads(MInst& inst,  mi& res, vd& exp, double alpha){
  //Check whether the single-end and paired-end reads support the existing results?
  //first, check single-end reads
  vi isvalid(inst.type.size(),1);
  for(int si=0;si<inst.type.size();si++){
    vi & csg=inst.type[si];//current read type
    for(int ii=0;ii<res.size();ii++){
      vi& resi=res[ii]; //current isoform discovered so far
      if(iscontained(csg,resi)){
        isvalid[si]=0;break;
      }
    }
  }
  //then, check paired-end reads
  for(int pi=0;pi<SUPPREADS.peindex.size();pi++){
    int pid=SUPPREADS.peindex[pi];
    int p1=inst.petype[pid].front();
    int p2=inst.petype[pid].back();
    if(isvalid[p1]==0 && isvalid[p2]==0){
      //should disable this type
      SUPPREADS.pecount[pi]=0;
    }
  }
  //check single-end reads
  for(int si=0;si<SUPPREADS.sgindex.size();si++){
    int sid=SUPPREADS.sgindex[si];
    if(isvalid[sid]==0 ){
      //should disable this type by decreasing the read count to 0
      SUPPREADS.sgcount[si]=0;
    }
  }

  //update SUPPREADS, remove unused paired-end and single-end records
  update_supporting_readinfo(inst);

  //this is the beginning of the weight index
  int index=3*SUPPREADS.nsegs+SUPPREADS.nijedges;
  //resize the weight index
  weights.resize(index+SUPPREADS.peindex.size()+SUPPREADS.sgindex.size());
  int totalr=0;
  //set up weights as the read counts * alpha
  for(int i=0;i<SUPPREADS.peindex.size();i++){
    weights[index]=alpha*SUPPREADS.pecount[i];
    totalr+=SUPPREADS.pecount[i]*2;
    index++;
  }
  for(int i=0;i<SUPPREADS.sgindex.size();i++){
    weights[index]=alpha/2*SUPPREADS.sgcount[i];
    totalr+=SUPPREADS.sgcount[i];
    index++;
  }
  
  return totalr;
}

/* Modify the original flow problem to remove unused read supporting variables and constratints */
int remove_unused_variables(glp_prob* lp){
  cout<<"Rows before removing:"<<glp_get_num_rows(lp)<<endl;
  cout<<"Columns before removing:"<<glp_get_num_cols(lp)<<endl;
  //this is the beginning of the weight index (column index)
  int index=3*SUPPREADS.nsegs+SUPPREADS.nijedges;//0-base index
  //set up weights as the read counts * alpha
  map<int,int> hasindex;
  vd weights2;
  for(int i=0;i<index;i++)
    weights2.push_back(weights[i]);
  vi peindex2,sgindex2;
  vi pecount2,sgcount2;
  for(int i=0;i<SUPPREADS.peindex.size();i++){
    if(weights[index]==0){
      hasindex[index]=1;//0-base index
      cout<<"Removing PE "<<SUPPREADS.peindex[i]<<", index:"<<index<<endl;
    }else{
      weights2.push_back(weights[index]);
      peindex2.push_back(SUPPREADS.peindex[i]);
      pecount2.push_back(SUPPREADS.pecount[i]);
    }
    index++;
  }
  for(int i=0;i<SUPPREADS.sgindex.size();i++){
    if(weights[index]==0){
      hasindex[index]=1;
      cout<<"Removing SG "<<SUPPREADS.sgindex[i]<<", index:"<<index<<endl;
    }else{
      weights2.push_back(weights[index]);
      sgindex2.push_back(SUPPREADS.sgindex[i]);
      sgcount2.push_back(SUPPREADS.sgcount[i]);
    }
    index++;
  }
  //update weights and other elements
  weights=weights2;
  SUPPREADS.peindex=peindex2;
  SUPPREADS.sgindex=sgindex2;
  SUPPREADS.pecount=pecount2;
  SUPPREADS.sgcount=sgcount2;
  //delete rows (constraints)
  map<int,int> rows_to_delete;
  
  int *p_ind=new int[glp_get_num_cols(lp)];
  double *p_val=new double [glp_get_num_cols(lp)];
  for(map<int,int>::iterator mit=hasindex.begin();
    mit!=hasindex.end();mit++){
    int t=mit->first;//1 base
    int len=glp_get_mat_col(lp,t+1,p_ind,p_val);
    for(int k=1;k<=len;k++){
      int rid=p_ind[k];//1-base
      if(p_val[k]!=0){
        rows_to_delete[rid]++; // 1-base
        cout<<"--deleting row "<<rid<<",val:"<<p_val[k]<<endl;
      }
    }
  }

  int *r_2_del=new int[rows_to_delete.size()+1];
  int r_2_del_index=1;
  for(map<int,int>::iterator mit=rows_to_delete.begin();
    mit!=rows_to_delete.end();mit++,r_2_del_index++) 
    r_2_del[r_2_del_index]=mit->first;//1-base
  if(r_2_del_index>1)
    glp_del_rows(lp,r_2_del_index-1,r_2_del);
  delete []r_2_del;

  //delete columns
  int* c_2_del=new int[hasindex.size()+1];
  int c_2_del_ind=1;
  for(map<int,int>::iterator mit=hasindex.begin();
    mit!=hasindex.end();
    mit++,c_2_del_ind++){
      c_2_del[c_2_del_ind]=mit->first+1;//convert to 1-base
      cout<<"--deleting column "<<c_2_del[c_2_del_ind]<<endl;
  }
  if(c_2_del_ind>1)
    glp_del_cols(lp,c_2_del_ind-1,c_2_del);
  delete []c_2_del;

  int lpnr=glp_get_num_rows(lp);
  int lpnc=glp_get_num_cols(lp);
  
  cout<<"Rows after removing:"<<lpnr<<endl;
  cout<<"Columns after removing:"<<lpnc<<endl;
  cout<<"weight length:"<<weights.size()<<endl;


  for(int t=0;t<lpnc;t++){
    int len=glp_get_mat_col(lp,t+1,p_ind,p_val);
    for(int k=1;k<=len;k++){
      int rid=p_ind[k];//1-base
      if(p_ind[k]>lpnr || p_ind[k]<=0){
        cout<<"Error: row index exceeds "<<lpnr<<endl;
      }
    }
  }


  delete [] p_ind;
  delete [] p_val;
  
  return 0;
}
/* Set up a proper GLPK linear programming problem.
   Now supporting read information is incorporated.
 */
int setupglpkprob(glp_prob* lp){
 
  //glp_erase_prob(lp);
  glp_set_prob_name(lp, "glpflow");
  glp_set_obj_dir(lp,GLP_MAX);


  return 0;
}
 

/* Set up a proper GLPK linear programming problem */
int setupglpkprob0(MInst& inst, glp_prob* lp,mi& rcount){
  glp_set_prob_name(lp, string(inst.id+"_glpflow").c_str());
  glp_set_obj_dir(lp,GLP_MAX);

  int nsegs=inst.getNSegs();
  cout<<"Instance "<<inst.id<<", nsegs:"<<nsegs<<endl;
  //for each nodes of the instance, select the number of its childrens.
  //This will determine the number of columns.
  int nijedges=0;
  for(int i=0;i<nsegs;i++){
    nijedges+=allchildren[i].size();
  }
 
  //In the original graph, there are 2N+2 nodes. They're marked as:
  // u_i: 0..N-1; 
  // v_i: N..2N-1; 
  // S: 2N; 
  // T: 2N+1
  //the number of flow constraints (rows) is 2N. They must be equal to 0
  glp_add_rows(lp,2*nsegs);
  //add flow constraints for u_i
  for(int i=0;i<nsegs;i++){
    stringstream ss;
    ss<<"r_u"<<i<<"_flowcons";
    string rn; ss>>rn;
    glp_set_row_name(lp,i+1,rn.c_str());
    glp_set_row_bnds(lp,i+1,GLP_FX,0.0,0.0);
  }
  //add flow constraints for v_i
  for(int i=0;i<nsegs;i++){
    stringstream ss;
    ss<<"r_v"<<i<<"_flowcons";
    string rn; ss>>rn;
    glp_set_row_name(lp,i+nsegs+1,rn.c_str());
    glp_set_row_bnds(lp,i+nsegs+1,GLP_FX,0.0,0.0);
  }
  //add variables (columns)
  //In the GLPK problem, the number of variables ( edges) is 3N+ (# childrens). They're marked as:
  //0..N-1: f(S, u_i), index in GLPK is 1..N
  //N..2N-1: f(v_i, T), index in GLPK is N+1..2N
  //2N..3N-1:f(u_i, v_i), index in GLPK is 2N+1..3N
  //the rest: f(v_i, u_j) 

  glp_add_cols(lp,3*nsegs+nijedges);
  int index=0;
  //add variable constraints for f(S,u_i)
  for(int i=0;i<nsegs;i++){
    stringstream ss;
    index=i;
    ss<<"c_S-u"<<i<<"_flow"; 
    string rn; ss>>rn;
    glp_set_col_name(lp,index+1,rn.c_str());
  }
   //add variable constraints for f(v_i,T)
  for(int i=0;i<nsegs;i++){
    stringstream ss;
    index=i+nsegs;
    ss<<"c_v"<<i<<"-T_flow"; 
    string rn; ss>>rn;
    glp_set_col_name(lp,index+1,rn.c_str());
  }
  //add variable constraints for f(u_i,v_i)
  for(int i=0;i<nsegs;i++){
    stringstream ss;
    index=i+nsegs*2;
    ss<<"c_v"<<i<<"-u"<<i<<"-T_flow"; 
    string rn; ss>>rn;
    glp_set_col_name(lp,index+1,rn.c_str());
  }
  //add variable constraints for f(v_i, u_j)
  for(int i=0;i<nsegs;i++){
    for(int j=0;j<allchildren[i].size();j++){
      index=allchildrenindex[i][j];
      stringstream ss;
      ss<<"c_v"<<i<<"-u"<<allchildren[i][j]<<"-T_flow"; 
      string rn; ss>>rn;
      glp_set_col_name(lp,index+1,rn.c_str());
    }
  }

  int* indarray=new int[3*nsegs];
  double* valarray=new double[3*nsegs];
  //add constraint matrix rows (still flow constraints)
  //u_i
  for(int i=0;i<nsegs;i++){
    //the number of edges to connected to u_i
    int indlen=2+allparent[i].size();
    // from S, u_i
    indarray[1]=i+1; valarray[1]=1.0; //incoming is positive
    // from u_i to v_i
    indarray[2]=2*nsegs+1+i; valarray[2]=-1.0;
    //for all edges starting from v_i
    for(int j=0;j<allparentindex[i].size();j++){
      index=allparentindex[i][j];
      indarray[3+j]=index+1; valarray[3+j]=1.0;
    }
    glp_set_mat_row(lp,i+1,indlen,indarray,valarray);
  }
  //v_i
  for(int i=0;i<nsegs;i++){
    //the number of edges to connected to u_i
    int indlen=2+allchildren[i].size();
    // from  v_i, T
    indarray[1]=nsegs+i+1; valarray[1]=-1.0;
    // from u_i to v_i
    indarray[2]=2*nsegs+1+i; valarray[2]=1.0;
    //for all edges starting from u_i
    for(int j=0;j<allchildrenindex[i].size();j++){
      indarray[3+j]=allchildrenindex[i][j]+1; valarray[3+j]=-1.0;
    }
    glp_set_mat_row(lp,nsegs+i+1,indlen,indarray,valarray);
  }
  delete[] indarray;
  delete[] valarray;

  return 0;
}


/* flow to isoform */
int flow2isoform(MInst& inst, glp_prob* lp, mi& res, mi& rcount){
  int nsegs=inst.getNSegs();
  //set up search index
  mi sindex(nsegs,vi(nsegs,-1));
  for(int i=0;i<allchildren.size();i++){
    for(int j=0;j<allchildren[i].size();j++){
      sindex[i][allchildren[i][j]]=allchildrenindex[i][j];
    }
  }
  //start and end flows
  int nstart=-1,nend=-1;
  vi thisisoform;
  mi addres;
  int ncurrent=-1;
  for(int i=0;i<nsegs;i++){
    double suflow=glp_get_col_prim(lp,i+1); // flows from S->u_i
    if(suflow>0){
      thisisoform.clear();
      ncurrent=i;
      while(ncurrent!=-1){
        thisisoform.push_back(ncurrent);
        double vtflow=glp_get_col_prim(lp,ncurrent+nsegs+1); //flows from v_i->T
        double uvflow=glp_get_col_prim(lp,i+2*nsegs+1); //flows from u_i->v_i
        if(uvflow<=0){cerr<<"Error: f(u_"<<ncurrent<<"->v_"<<ncurrent<<")=0\n";}
        if(vtflow>0){
          ncurrent=-1;
          addres.push_back(thisisoform);
          break;
        }
        int nbranch=0;
        int nnext=-1;
        double nmaxweight=-1000000.0;
        for(int j=0;j<allchildren[ncurrent].size();j++){
           int cidx=allchildrenindex[ncurrent][j]+1;
           double ijflow=glp_get_col_prim(lp,cidx);
           double ijweight=glp_get_obj_coef(lp,cidx);
           if(ijflow>0){
             nbranch++;
             if(nbranch>1) 
               cerr<<"Warning: branches for node "<<ncurrent<<endl;
             if(ijweight>nmaxweight){
               //in cases of branches, choose the one with maximum weight
               nnext=allchildren[ncurrent][j];
               nmaxweight=ijweight;
             }
           }
        }//end for
        ncurrent=nnext;
      }//end while
    }//end if suflow>0
  }
  //the flow is zero, no isoforms added
  if(addres.size()==0)return -1; 
  int nadded=0;
  for(int j=0;j<addres.size();j++){
    thisisoform=addres[j];
    //check the junction flows
    for(int i=1;i<thisisoform.size();i++){
      int p1=thisisoform[i-1],p2=thisisoform[i];
      int did=sindex[p1][p2];
      if(did==-1){
        cerr<<"Error: -1 index for nodes u_"<<p1<<" and v_"<<p2<<endl;
      }else{
        double uvflow=glp_get_col_prim(lp,did+1); //flows from v_i->T
        if(uvflow==0){
          cerr<<"Error: 0 flow for nodes u_"<<p1<<" and v_"<<p2<<endl;
        }else{
        }
      }
    }
    //check if there are existing isoforms identical to this one?
    bool hasidentical=false;
    for(int k=0;k<res.size();k++){
      if(thisisoform==res[k]){
        hasidentical=true;
        break;
      }
    }
    if(hasidentical==false){
      cout<<"Isoform: ";for(int i=0;i<thisisoform.size();i++)cout<<thisisoform[i]<<",";cout<<endl;
      res.push_back(thisisoform);
      nadded++;
    }
  }
  return nadded;
}

/* Print the flow problem */
void printflowproblem(MInst& inst, glp_prob* lp){
  
  bool print_read_support=true;
  int nsegs=inst.getNSegs();
  int index;
  
  cout<<"Flow problem:\n";
  for(int i=0;i<nsegs;i++){
    double wsu=glp_get_obj_coef(lp,i+1);
    double wvt=glp_get_obj_coef(lp,i+nsegs+1);
    double wvu=glp_get_obj_coef(lp,i+2*nsegs+1);
    cout<<i<<"\tw(S,i)="<<wsu<<"\tw(i,T)="<<wvt<<"\tw(u_i,v_i)="<<wvu;
    cout<<"\tChildren:";
    for(int j=0;j<allchildren[i].size();j++){
      index=allchildrenindex[i][j];
      double wij=glp_get_obj_coef(lp,index+1);
      cout<<allchildren[i][j]<<"("<<wij<<",i:"<<index<<"),";
    }
    cout<<endl;
  }
  if(print_read_support){
    cout<<"Reads support:\n";
    for(int x=0;x<SUPPREADS.peindex.size();x++){
      int petypeid=SUPPREADS.peindex[x];
      int p1=inst.petype[petypeid].front();
      int p2=inst.petype[petypeid].back();
      double wp=glp_get_obj_coef(lp,3*nsegs+SUPPREADS.nijedges+1+x);
      cout<<"PEType: "<<petypeid<<"("<<p1<<"[";
      for(int pj=0;pj<inst.type[p1].size();pj++)cout<<inst.type[p1][pj]<<",";
      cout<<"],"<<p2<<"[";
      for(int pj=0;pj<inst.type[p2].size();pj++)cout<<inst.type[p2][pj]<<",";
      cout<<"])\tr:"<<getVectorSum<int>(inst.petypecount[petypeid])<<"\tweight:"<<wp<<endl;
    }
    for(int x=0;x<SUPPREADS.sgindex.size();x++){
      int sgtypeid=SUPPREADS.sgindex[x];
      double wp=glp_get_obj_coef(lp,3*nsegs+SUPPREADS.nijedges+SUPPREADS.peindex.size()+1+x);
      cout<<"SGType: "<<sgtypeid<<"[";
      for(int pj=0;pj<inst.type[sgtypeid].size();pj++)cout<<inst.type[sgtypeid][pj]<<",";
      cout<<"]\tr:"<<getVectorSum<int>(inst.typecount[sgtypeid])<<"\tweight:"<<wp<<endl;
    }
  }
  
}

string vec2str(vector<int>& v){
  stringstream ss;
  for(int i=0;i<v.size();i++){
    ss<<v[i];
    if(i<v.size()-1)ss<<",";
  }
  return ss.str();
}
/* Print flow for debugging */
void printflow(MInst& inst, glp_prob* lp){
  bool print_read_support=true;
  int nsegs=inst.getNSegs();
  int index;
  for(int i=0;i<nsegs;i++){
    double suflow=glp_get_col_prim(lp,i+1); // flows from S->u_i
    double vtflow=glp_get_col_prim(lp,i+nsegs+1); //flows from v_i->T
    double vuflow=glp_get_col_prim(lp,i+2*nsegs+1); //flows from v_i->T
    cout<<i<<"\tf(S,i):"<<suflow<<"\tf(i,T):"<<vtflow<<"\tf(u_i,v_i):"<<vuflow;
    cout<<"\tC:";
    for(int j=0;j<allchildren[i].size();j++){
      index=allchildrenindex[i][j];
      double ijflow=glp_get_col_prim(lp,index+1);
      cout<<allchildren[i][j]<<"("<<ijflow<<"),";
    }
    cout<<endl;
  }
  if(print_read_support){
    cout<<"Reads support:\n";
    for(int x=0;x<SUPPREADS.peindex.size();x++){
      int petypeid=SUPPREADS.peindex[x];
      int p1=inst.petype[petypeid].front();
      int p2=inst.petype[petypeid].back();
      double wp=glp_get_col_prim(lp,3*nsegs+SUPPREADS.nijedges+1+x);
      double wp0=glp_get_obj_coef(lp,3*nsegs+SUPPREADS.nijedges+1+x);
      cout<<"PEType: "<<petypeid<<"\t";
      stringstream ss0;
      ss0<<p1<<"[";
      ss0<<vec2str(inst.type[p1]);
      ss0<<"],"<<p2<<"[";
      ss0<<vec2str(inst.type[p2]);
      ss0<<"]";
      cout<<setw(20)<<ss0.str();
      cout<<"\tr:"<<getVectorSum<int>(inst.petypecount[petypeid]);
      cout<<"\tw:"<<setw(7)<<wp0;
      cout<<"\tflow:"<<wp<<endl;
    }
    for(int x=0;x<SUPPREADS.sgindex.size();x++){
      int sgtypeid=SUPPREADS.sgindex[x];
      double wp=glp_get_col_prim(lp,3*nsegs+SUPPREADS.nijedges+SUPPREADS.peindex.size()+1+x);
      double wp0=glp_get_obj_coef(lp,3*nsegs+SUPPREADS.nijedges+SUPPREADS.peindex.size()+1+x);
      cout<<"SGType: "<<sgtypeid<<"\t[";
      cout<<setw(16)<<vec2str(inst.type[sgtypeid]);
      cout<<"]\tr:"<<getVectorSum<int>(inst.typecount[sgtypeid]);
      cout<<"\tw:"<<setw(7)<<wp0;
      cout<<"\tflow:"<<wp<<endl;
    }
  }
 
}

/* set up children and parent information */
void setupconn(MInst& inst,mi& rcount,EnumParameter para){
  int nsegs=inst.getNSegs();
  allchildren=mi(nsegs);
  allchildrenindex=mi(nsegs);
  allparent=mi(nsegs);
  allparentindex=mi(nsegs);
  int index=3*nsegs;
  for(int i=0;i<nsegs;i++){
    selectchildren(inst,i,rcount, allchildren[i],para);
    for(int j=0;j<allchildren[i].size();j++){
      cout<<i<<"->"<<allchildren[i][j]<<",";
      allchildrenindex[i].push_back(index);
      allparent[allchildren[i][j]].push_back(i);
      allparentindex[allchildren[i][j]].push_back(index);
      index++;
    }
    cout<<endl;
  }

  //check if each segment is in the introns of the other segments?
  vi isintron(nsegs,0);
  for(int i=0;i<rcount.size();i++){
    int prevint=-1;
    for(int j=i+1;j<rcount[i].size();j++){
      if(rcount[i][j]>0){
        if(j!=prevint && prevint!=-1) isintron[prevint]=1;
        prevint=j;
      }
    }
  }
  cout<<"ISINTRON:"; for(int i=0;i<isintron.size();i++) if(isintron[i]==1)cout<<i<<" ";cout<<endl;
  

  //check if there are missing links
  //disabled: since it will most likely connect links dropped by selectchildren
  for(int ii=0;ii<nsegs-1 ;ii++){
    if(allchildren[ii].size()==0 && allparent[ii+1].size()==0
       && isintron[ii]==0 && isintron[ii+1]==0
      ){
      //check paired-end reads to see if they can be connected
      for(int pp=0;pp<inst.petype.size();pp++){ 
        int p0=inst.petype[pp].front();
        int p1=inst.petype[pp].back();
        if(p0!=p1 && inst.type[p0].back()==ii 
          && inst.type[p1].front()==ii+1 && getVectorSum<int>(inst.petypecount[pp])>0){
           cout<<"Manually connecting segment "<<ii<<" and "<<ii+1<<"\n";
           allchildren[ii].push_back(ii+1);
           allparent[ii+1].push_back(ii);
           allchildrenindex[ii].push_back(index);
           allparentindex[ii+1].push_back(index);
           index++;
           break;
        }
      }
    }
  }
 
}

/* Preparation of flow problem. It sets up the flow problem */
void enumiso_prepareflow(MInst& inst,mi& rcount,int dir,EnumParameter para){
  // fill rcount
  inst.fillRCount(rcount,dir);
  //record the selected childrens and parents
  setupconn(inst,rcount,para);

  getcongraph(inst,allparent,allconn);
  // create a GLPK problem
  //glp_prob *lp=glp_create_prob();
  // set up graphs

  //determine the number of extra variables (t_r)
  getsupportingreadinfo(inst,para);
  //if you want to disable read supporting variables:
  //SUPPREADS.clear();
  //SUPPREADS.nvars=SUPPREADS.nconstraints=0;
  //return lp;

  
  //check the correlation among all segments
  if(para.DO_CORRELATION)
    setup_correlations(inst,rcount,para);
}

/* The main interface of using flow to predict isoforms
   Notice that res and exp may not be empty (it may include existing isoforms)
   Return value: the new number of isoforms added to res; or -1 if the flow produces no isoforms.
*/
int enumiso_flow(MInst& inst, glp_prob* lp, mi& rcount, mi & res, vd& exp,EnumParameter para){ 
  cout<<"Begin enumeration...\n";

  setupglpkprob(lp);
  // set up weights and bounds
  updateboundandweight(inst,  lp, rcount,res,exp,para);
  //solve the problem
  printflowproblem(inst,lp);
  int retval=glp_simplex(lp,NULL);
  if(retval!=0){// 
    cerr<<"Error: ret code "<<retval<<endl;
  }
  printflow(inst,lp);
  //retrieve the flow values
  int nadded=flow2isoform(inst,  lp,  res, rcount );
  return nadded;
}
/* remove duplications */
int removedup(MInst& inst,mi& res,md&explv,vi& direction){
  vi valid(res.size(),1);
  //pairwise checking, filter identical and contain case
  cout<<"Candidate:\n";
  for(int i=0;i<res.size();i++){
    cout<<i<<":";
    for(int j=0;j<res[i].size();j++)cout<<res[i][j]<<" ";cout<<endl;
  }
  for(int i=0;i<res.size();i++){
    cout<<"Checking "<<i<<endl;
    if(valid[i]==0)continue;
    for(int j=i+1;j<res.size();j++){
      if(valid[j]==0)continue;
      if(res[j]==res[i]){
        valid[j]=0;
        cout<<i<<" is identical to "<<j<<endl;
        continue;//break;
      }
      //check if i is included in j, or if j is included in i
      if(iscontained(res[i],res[j])){
        cout<<i<<" in "<<j<<endl;
        valid[i]=0;
        continue;
      }
      if(iscontained(res[j],res[i])){
        cout<<j<<" in "<<i<<endl;
        valid[j]=0;
        //break;
      }
    }
  }
  mi resb=res;
  md explvb=explv;
  vi dirb=direction;
  res.clear();explv.clear();direction.clear();
  for(int i=0;i<resb.size();i++){
    if(valid[i]==1){
      res.push_back(resb[i]);
      explv.push_back(explvb[i]);
      direction.push_back(dirb[i]);
    }
  }
  return 0;

}

/* filter isoforms */
int  filterisoforms(MInst& inst,mi& res,md&explv,vi& direction){
  vi valid(res.size(),1);
  vd exp(res.size());
  vd avgp0(inst.getNSegs());
  for(int i=0;i<inst.getNSegs();i++)
    avgp0[i]=getVectorSum<double>(inst.segp0[i])/inst.segp0[i].size();
  for(int i=0;i<res.size();i++){
    exp[i]=getVectorSum<double>(explv[i])/explv[i].size();
  }
  //for single-exon ones, set up higher standards for them 
  for(int i=0;i<res.size();i++){
    bool iscontinuous=true;
    for(int j=0;j<res[i].size()-1;j++){
      if(res[i][j+1]!=res[i][j]+1){
        iscontinuous=false;
        break;
      }
    }
    if(iscontinuous){
      //check the exp lv
      //if(exp[i]<EXPCUTOFF)valid[i]=0;
      //vd allp0(res[i].size());
      //for(int j=0;j<res[i].size();j++) allp0[j]=avgp0[res[i][j]];
      //double aap0=getVectorSum<double>(allp0)/allp0.size();
      //if(aap0<SEGP0CUTOFF)valid[i]=0;
      //check if it is contained within the other one?
      long lefti=inst.segment[res[i][0]][0];
      long righti=inst.segment[res[i].back()][1];
      //check all other ones
      for(int j=0;j<res.size();j++){
        if(j==i)continue;
        long leftj=inst.segment[res[j][0]][0];
        long rightj=inst.segment[res[j].back()][1];
        if(lefti>leftj && righti<rightj){
          if(exp[i]<EXPCUTOFF){ valid[i]=0;break;}
        }
      }
    }
  }

  mi resb=res;
  md explvb=explv;
  vi dirb=direction;
  res.clear();explv.clear();direction.clear();
  for(int i=0;i<resb.size();i++){
    if(valid[i]==1){
      res.push_back(resb[i]);
      explv.push_back(explvb[i]);
      direction.push_back(dirb[i]);
    }
  }
  return 0;
}

int doflow_together(MInst& inst, mi& res, vi& direction, EnumParameter para ){
  mi rcount;
  //enumerate two possible directions
  glp_prob *lp=glp_create_prob();
  for(int dir=0;dir>=-1;dir-=2){//may create some problems here. fix it later
    cout<<"Flow of direction "<<dir<<"...\n";
    rcount.clear();
    enumiso_prepareflow(inst,rcount,dir,para);
    mi res0;
    vd exp0;
    int niter=0;
    double maxobjval=-100000.0;
    while(true){
      glp_erase_prob(lp);
      //use flow algorithm
      int nadded=enumiso_flow(inst,lp,rcount,res0,exp0,para);
      //predict expression level
      cout<<"# of predicted isoforms: "<<res.size()<<",added isoforms:"<<nadded<<endl;
      exp0=vd(res0.size(),1.0);
      if(nadded==-1)break;
      niter++;
      //break if exceeds the maximum number of iteration
      if(niter>100)break;
      //break if the objective function is too small
      double currentobjval=glp_get_obj_val(lp);
      if(currentobjval>maxobjval)maxobjval=currentobjval;
      if(currentobjval<MAX_OBJ_FRACTION*maxobjval)break;
    }
    weights.clear();bounds.clear();
    //add direction
    for(int j=0;j<res0.size();j++)res.push_back(res0[j]);
    for(int j=direction.size();j<res.size();j++){
      direction.push_back(inst.dir);
    }
  }
  glp_delete_prob(lp);

  return 0;
}


int doflow_bysample(MInst& inst, mi& res, vi& direction, EnumParameter para ){
  mi rcount;
  //enumerate two possible directions
  if(inst.getNSamples()<=0)return 0;
  //If there is only one segment, return directly.
  if(inst.getNSegs()==1){
    res.push_back(vi(1,0));
    direction.push_back(inst.dir);
    return 0;
  }
  // create a GLPK problem
  glp_prob *lpp=glp_create_prob();
  for(int ns=0;ns<inst.getNSamples();ns++){
    cout<<"###SAMPLE "<<ns<<endl;
    para.current_sample=ns;
    rcount.clear();
    enumiso_prepareflow(inst,rcount,0,para);
    mi res0;
    vd exp0;
    int niter=0;
    double maxobjval=-100000.0;
    rcount.clear();
    inst.fillRCount(rcount,0,ns);
    while(true){
      //use flow algorithm
      glp_erase_prob(lpp);
      int nadded=enumiso_flow(inst,lpp,rcount,res0,exp0,para);
      //predict expression level
      cout<<"# of predicted isoforms: "<<res.size()<<",added isoforms:"<<nadded<<endl;
      exp0=vd(res0.size(),1.0);
      if(nadded==-1)break;
      niter++;
      //break if exceeds the maximum number of iteration
      if(niter>2)break;
      //break if the objective function is too small
      double currentobjval=glp_get_obj_val(lpp);
      if(currentobjval>maxobjval)maxobjval=currentobjval;
      if(currentobjval<MAX_OBJ_FRACTION*maxobjval)break;
    }
    weights.clear();bounds.clear();
    //add direction
    for(int j=0;j<res0.size();j++)res.push_back(res0[j]);
    for(int j=direction.size();j<res.size();j++){
      direction.push_back(inst.dir);
    }
  }
  glp_delete_prob(lpp);

  return 0;
}

/* Using flow algorithm to find isoforms, and calculate their expression levels */
int flowandcalexp(MInst& inst, mi& res, md& explv,vi& direction,EnumParameter para){

  vd exp;

  //execute the flow algorithm
  if(!para.assemble_by_sample)
    doflow_together(inst, res, direction, para );
  else
    doflow_bysample(inst, res, direction, para );

  explv.resize(res.size());
  removedup(inst,res,explv,direction);
  cout<<"First calculation...\n";
  calcexplv(inst, res, explv);
  //filter results
  filterisoforms(inst,res,explv,direction);
  calcexplv(inst, res, explv);
  
  exp.resize(res.size());
  //calculate the mean value
  for(int i=0;i<res.size();i++){
    exp[i]=getVectorSum<double>(explv[i])/explv[i].size();
  }
  //print results
  cout<<"## Results:\n";
  for(int i=0;i<res.size();i++){
    cout.precision(2);
    for(int j=0;j<explv[i].size();j++) {cout<<explv[i][j]<<" ";} cout<<"\t";
    for(int j=0;j<res[i].size();j++) cout<<res[i][j]<<" "; 
    cout<<(direction[i]>0?'+':'-');
    cout<<endl;
  }

  return 0;
}

