#include <cstdlib>
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <sys/stat.h>
#include <pthread.h>

using namespace std;

int nthread=1;

// parameters

/* input bam files */
vector<string> inputfiles;
/* the prefix for processsam program */
vector<string> prefixes;
/* sample names */
vector<string> labels;

/* range passed to processsam */
string samrange=" ";

/* processsam option for 1st pass and 2nd pass */
string processsam1stoption=" -g 10 -k 100000 -c 4 -i -a ";
string processsam2ndoption=" -g 10 -k 100000 -c 4 -i ";


/* predminst options */
string predminstoptions=" --assemble-by-sample --min-frac 0.05 --no-correlation ";

/* Use COMMAND instead of BAM file. Controlled by -c parameter */
bool USE_COMMAND=false;

/* The input is SAM file instead of BAM file */
bool USE_SAM=false;

/* temp dir. Controlled by -t parameter */
string tmpdir="tmp/";

/* string to store the number of samples */
string nsamplestr="";

/*debugging commands, represent steps 1,2,3,4,5,6. Controlled by --debug parameter */
bool firstpass=true;
bool merge1stpass=true;
bool secondpass=true;
bool merge2ndpass=true;
bool flowpass=true;
bool postpass=true;

bool supressoutput=true;

/* used by each pthread to check if each job is completed or not */
int* pt_completed;
/* mutex for accessing pt_completed */
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
/* mutex for accessing inputfiles and prefixes */
pthread_mutex_t mutex2 = PTHREAD_MUTEX_INITIALIZER;

/* thread info, passed to thread function */
struct threadinfo{
  int jobid;
  int threadid;
  int ntotalthread;
};

/* Intermediate file names */
string objbound;
string mergedbed;
string alljunc;
string finalinstfile;
string finalinstpred;
string finalinstpredexp;


/* Function declaration */
int execpthread(int jobid);

/* Print help message */
void printmsg(){
  cout<<"Usage: runminst {OPTIONS} <BAM file 1|COMMAND 1> <BAM file 2|COMMAND 2> ...\n\n";
  cout<<"Options:\n\n";
  cout<<"\t-h/--help\t\tPring help message.\n\n";
  cout<<"\t-r/--range <range>\tSpecify the range of the bam file. The ranges are specified as 'chrname:start-end'; if multiple ranges are specified, they must be separated by comma (,).\n\n";
  cout<<"\t-t/--tmp <string>\tSpecify the temporary dir. Default tmp/\n\n";
  cout<<"\t-p/--pthread <int>\tSpecify the number of threads used. Default 1.\n\n";
  cout<<"\t--debug <int,...>\tSpecify the steps executed. The integer value must be between 1-6. Default 1,2,3,4,5,6.\n\n";
  cout<<"\t-c/--command\t\tInstead of providing BAM file names, use STDOUT of commands executable under the current path.\n";
  cout<<"\t-s/--sam\t\tThe file format is SAM instead of BAM. If it is on, -r/--range option is not allowed.\n";
  cout<<"\t-L/--label <string,...>\tSpecify the labels of the input files, separated by comma. The number of labels must equal to the input files. Default \"sample_n\".\n\n";
}

/* Parse options */
int parseopt(int argc,char* argv[]){
  int ii=1;
  while(ii<argc){
    string cu(argv[ii]);
    if(cu=="-h" || cu=="--help"){
      printmsg();
      return -1;
    } 
    else if(cu=="-r" || cu=="--range" ){
      samrange=string(argv[ii+1]);
      for(int i=0;i<samrange.size();i++)if(samrange[i]==',')samrange[i]=' ';
      ii++;
    } 
    else if(cu=="-p" || cu=="--pthread" ){
      nthread=atoi(argv[ii+1]);
      cout<<"Using "<<nthread<<" threads.\n";
      ii++;
    }
    else if(cu=="-t" || cu=="--tmp" ){
      tmpdir=string(argv[ii+1]);
      if(tmpdir[tmpdir.size()-1]!='/') tmpdir+="/";
      ii++;
    }
    else if(cu=="-L" || cu=="--label"){
      string alllabels=string(argv[ii+1]);
      //split by ','
      int nst=-1;
      while((nst=alllabels.find(','))!=-1){
        labels.push_back(alllabels.substr(0,nst));
        alllabels=alllabels.substr(nst+1);
      }
      if(alllabels.size()>0) labels.push_back(alllabels);
      ii++;
    }
    else if(cu=="-c" || cu=="--command"){
      USE_COMMAND=true;
    }
    else if(cu=="-s" || cu=="--sam"){
      USE_SAM=true;
    }
    else if(cu=="--debug" ){
      string steps(argv[ii+1]);
      stringstream ss(steps);
      firstpass=merge1stpass=secondpass=merge2ndpass=flowpass=postpass=false;
      int x;char c;
      
      while(ss.good()){
         ss>>x;
         ss>>c;
         switch(x){
           case 1: firstpass=true; break;
           case 2: merge1stpass=true; break;
           case 3: secondpass=true; break;
           case 4: merge2ndpass=true; break;
           case 5: flowpass=true; break;
           case 6: postpass=true; break;
           default: break;
         }
      }
      ii++;
    }
    else{
      inputfiles.push_back(string(argv[ii]));
    }
    ii++;
  }
  prefixes.resize(inputfiles.size());

  if(labels.size()>0){cout<<"Labels: "; for(int i=0;i<labels.size();i++)cout<<labels[i]<<" ";cout<<endl;}
  if(labels.size()>0 && labels.size()!=inputfiles.size()){
    cerr<<"Error: the label size ("<<labels.size()<<") must be equal to the number of input files ("<<inputfiles.size()<<").\n";
    return -1;
  }
  return 0;
}

/* Determine the names of the intermediate files
*/
void determineintermediatefilename(){
  stringstream ss;
  ss<<inputfiles.size();
  ss>>nsamplestr;
  objbound=nsamplestr;
  mergedbed=objbound;
  alljunc=objbound;
  finalinstfile=objbound;

  objbound=tmpdir+"allbound."+objbound+".bed";
  mergedbed=tmpdir+"merged."+mergedbed+".bed";
  alljunc=tmpdir+"alljunc."+alljunc+".txt";
  finalinstfile=tmpdir+"allpred."+finalinstfile+".ins";
  finalinstpred=finalinstfile+".pred";
  finalinstpredexp=finalinstfile+".pred.explv";
}


/*
Check prerequisite programs
*/
int checkprerequisite(){
  if(!system(NULL)){
    cerr<<"Error: cannot execute system command in std C++.\n";
    return -1;
  }
  if(system("cat --version > /dev/null")!=0)return -1;
  if(system("sort --version > /dev/null")!=0)return -1;
  int c;
  c=system("mergerange");
  if(c!=0)return -1;
  c=system("manalyzecvg");
  cout<<c<<endl;
  if(c!=0)return -1;
  c=system("processsam 2>/dev/null");
  if(c!=0)return -1;
  c=system("samtools 2> /dev/null");
  if(c>256)return -1;
  struct stat sb;
  if(stat(tmpdir.c_str(),&sb)==0 && S_ISDIR(sb.st_mode) ){
      cout<<"tmp dir "<<tmpdir<<" exists.\n";
  } else{
      string cmdmkdir="mkdir "+tmpdir;
      cout<<"Creating tmp dir "<<tmpdir<<"...\n";
      if(system(cmdmkdir.c_str())!=0){
        cerr<<"Error when creating directories.\n";
        exit(-1);
      }
  }
  return 0;
}

/* Running system command */
int runcommand(string cmd){
  int c=0;
  cout<<"cmd:"<<cmd<<endl<<flush;
  c=system(cmd.c_str());
  cout<<"cmd:"<<cmd<<",return:"<<c<<endl<<flush;
  return 0;
}

/* First pass of the BAM file: generate individual ranges and coverages */
int runfirstpass(string bamfile,string prefix){
  string samcommand="samtools view "+bamfile+" "+samrange+" ";
  if(USE_COMMAND){
    samcommand=bamfile;
  }
  if(USE_SAM){
    samcommand="cat "+bamfile+" ";
  }
  string outprefix=prefix;
  string processsamcmd=" processsam "+processsam1stoption+ " -o "+ outprefix + " - ";

  string command=samcommand+" | "+processsamcmd;
  if(supressoutput)command+=" 1>/dev/null";
  
  runcommand(command);
  return 0;
}
/* Step 2: merge ranges and analyze coverage files */
int mergerangeandcvg(){
  //merge range and cvg file
  string cmdcatpg="";
  string cmdwigpg="";
  
  cout<<"-----------2:merging ranges and junctions-------------\n";

  for(int i=0;i<inputfiles.size();i++){
    string prefix=tmpdir+labels[i]+".1st.bound.bed ";
    string wigprefix=tmpdir+labels[i]+".1st.real.wig ";
    cmdcatpg+=prefix;
    cmdwigpg+=wigprefix;
  }
  cmdcatpg="cat "+cmdcatpg+" | sort -k 1,1 -k 2,2n > "+objbound;
  runcommand(cmdcatpg);
  
  string cmdmergerange="mergerange "+objbound+" > "+mergedbed;
  runcommand(cmdmergerange);

  cmdwigpg="manalyzecvg "+cmdwigpg+" | sort -k 1,1 -k 2,2n > "+alljunc;
  runcommand(cmdwigpg);
  return 0;
}

/*
Step 3: second pass of the BAM file. Use given gene range and boundary
*/
int runsecondpass(string bamfile,string prefix){
  string samcommand="samtools view "+bamfile+" "+samrange+" ";
  if(USE_COMMAND){
    samcommand=bamfile;
  }
  if(USE_SAM){
    samcommand="cat "+bamfile+" ";
  }
  string outprefix=prefix;
  string processsamcmd=" processsam  -r  "+mergedbed +" -e "+alljunc +processsam2ndoption+ " -o "+ outprefix + " - ";

  string command=samcommand+" | "+processsamcmd;
  if(supressoutput)command+=" 1>/dev/null";
  
  runcommand(command);
  return 0;
}

/*
Step 4: merge instances and create individual instances 
*/
int mergeinstandrunpredminst(){
  
  cout<<"-----------4:merging instances-------------\n";
  string mergeinst=" ";
  for(int i=0;i<inputfiles.size();i++){
    string instfile=tmpdir+labels[i]+".2nd."+nsamplestr+".instance";
    mergeinst+=instfile+" ";
  }

  string cmdmergeinstances="mergeinstances -o "+finalinstfile+" "+mergeinst;
  runcommand(cmdmergeinstances);
  return 0;
}


/* 
  Step 5: predicting 
*/
int runflow(){
  cout<<"-----------5:predicting-------------\n";
  //old code
  //string cmdpredminst="predminst -o "+finalinstpred+" "+" -e "+alljunc+" "+finalinstfile;
  //runcommand(cmdpredminst);

  int jobid=3;
  execpthread(jobid);
  
  string cmdcatpred="cat ";
  string cmdcatpredexp="cat ";
  //concaternate predictions from individual threads
  for(int i=0;i<nthread;i++){
    stringstream sst,sst2;
    sst<<tmpdir<<"pthreadpred_"<<i<<"_"<<nthread<<".pred ";
    sst2<<tmpdir<<"pthreadpred_"<<i<<"_"<<nthread<<".pred.explv ";
    cmdcatpred+=sst.str();
    cmdcatpredexp+=sst2.str();
  }
  cmdcatpred+=" > "+finalinstpred;
  cmdcatpredexp+=" > "+finalinstpredexp;
  runcommand(cmdcatpred);
  runcommand(cmdcatpredexp);
  return 0;
}



/*
Step 6: Post processing 
*/
int postprocess(){
  cout<<"-----------6:post processing-------------\n";
  string cmdpred2gtf="pred2gtf -o - "+finalinstpred+" | sortgtf -o "+finalinstpred+".gtf -";
  runcommand(cmdpred2gtf);
  //string cmdsortgtf="sortgtf "+finalinstpred+".gtf";
  //runcommand(cmdsortgtf);
  return 0;
}

/* Multithread entry 
Jobid:
	1 processsam first pass;
	2 processsam second pass;
	3 predminst thread
*/
void* run_pthread(void* xx){
  threadinfo* pt=(threadinfo*)xx;
  int jobid=pt->jobid;
  int threadid=pt->threadid;
  int ntotalthread=pt->ntotalthread;
  if(jobid==1 || jobid==2){
    //processsam jobs
    string cinputfiles="";
    string cprefix="";
    while(true){
       bool hasjob=false;
       pthread_mutex_lock( &mutex1 );
       for(int i=0;i<inputfiles.size();i++){
         if(pt_completed[i]==0){
           pt_completed[i]=1;
           cinputfiles=inputfiles[i];
           cprefix=prefixes[i];
           hasjob=true;
           break;
         }
       }
       pthread_mutex_unlock( &mutex1 );
       
       if(hasjob==false)break;
       if(jobid==1){
         pthread_mutex_lock( &mutex2 );
         cout<<"1st pass "<<cinputfiles<<", p="<<threadid<<"\n";
         pthread_mutex_unlock( &mutex2 );
         runfirstpass(cinputfiles,cprefix);
       }else if(jobid==2){
         pthread_mutex_lock( &mutex2 );
         cout<<"2nd pass "<<cinputfiles<<", p="<<threadid<<"\n";
         pthread_mutex_unlock( &mutex2 );
         runsecondpass(cinputfiles,cprefix);
       }else{
         cerr<<"Error: unknown job id.\n";
       }
    }
  }
  if(jobid==3){
    cout<<"-----------5:predicting-------------\n";
    string cmdpredminst="predminst ";// -e "+alljunc;
    stringstream sst;
    sst<<" -p "<<threadid<<","<<ntotalthread<<" ";
    sst<<" -o "<<tmpdir<<"pthreadpred_"<<threadid<<"_"<<ntotalthread<<".pred ";
    //sst<<" --assemble-by-sample ";
    sst<<predminstoptions;
    cmdpredminst+=sst.str();
    cmdpredminst+=" "+finalinstfile;
    if(supressoutput)cmdpredminst+=" 1>/dev/null";
    runcommand(cmdpredminst);
  }
  cout<<"Thread "<<threadid<<" exit...\n";
  return 0;
}

int execpthread(int jobid){
  pt_completed=new int[inputfiles.size()];
  threadinfo* allthif=new threadinfo[nthread];
  for(int i=0;i<inputfiles.size();i++)pt_completed[i]=0;

  pthread_t* thread_id=new pthread_t[nthread];
  for(int i=0;i<nthread;i++){
    allthif[i].jobid=jobid; allthif[i].threadid=i; allthif[i].ntotalthread=nthread;
    pthread_create( &thread_id[i], NULL, run_pthread, (void*)(&allthif[i]));
  }
  for(int i=0;i<nthread;i++)
      pthread_join( thread_id[i], NULL);

  delete []allthif;
  delete []thread_id;
  delete []pt_completed;
}

/* Main Entry */
int main(int argc, char* argv[]){
  if(argc==1){
    printmsg();
    return 0;
  }
  if(parseopt(argc,argv)==-1)return -1;
  //check prerequisite
  if(checkprerequisite()==-1) {
    cerr<<"Prerequiste check failed. Check your system to see if the required programs are installed (like samtools).\n";
    return -1;
  }
  //labels
  if(labels.size()==0){
    for(int i=0;i<inputfiles.size();i++) {
      stringstream ss;
      ss<<"sample_"<<i;
      string ck;
      ss>>ck;
      labels.push_back(ck);
    }
  }
  cout<<"Input files:\n";
  for(int i=0;i<inputfiles.size();i++) {
    cout<<inputfiles[i]<<endl;
  }

  determineintermediatefilename();
  //Step 1: run 1st pass  
  if(firstpass){
    cout<<"-----------1: first pass-------------\n";
    for(int i=0;i<inputfiles.size();i++){
      string prefix=tmpdir+labels[i]+".1st";
      prefixes[i]=prefix;
      //runfirstpass(inputfiles[i],prefix);
    }
    int jobid=1;
    execpthread(jobid);
  }
  //Step 2: merge 1st pass results
  if(merge1stpass) mergerangeandcvg();
  //Step 3: run 2nd pass  
  if(secondpass){
  cout<<"-----------3: second pass-------------\n";
    for(int i=0;i<inputfiles.size();i++){
      string prefix=tmpdir+labels[i]+".2nd."+nsamplestr;
      prefixes[i]=prefix;
      //runsecondpass(inputfiles[i],prefix);
    }
    int jobid=2;
    execpthread(jobid);
  }
  //Step 4: merge instance and run predminst
  if(merge2ndpass) mergeinstandrunpredminst();
  
  //Step 5: predicting
  if(flowpass) runflow();
  //Step 6: post processing
  if(postpass)postprocess();

  return 0;
}
